from mpfmc.uix.display import Display
from mpfmc.uix.slide import Slide
from mpfmc.tests.MpfMcTestCase import MpfMcTestCase


class TestDisplaySingle(MpfMcTestCase):
    def get_machine_path(self):
        return 'tests/machine_files/display'

    def get_config_file(self):
        return 'test_display_single.yaml'

    def test_mc_display(self):
        # Make sure a single display is loaded properly:

        self.assertIn('window', self.mc.displays)
        self.assertTrue(isinstance(self.mc.displays['window'], Display))
        self.assertEqual(self.mc.displays['window'].size, [401, 301])
        self.assertEqual(self.mc.targets['default'], self.mc.targets[
            'window'])

        # walk the display's widget tree and make sure everything is right
        widget_hierarchy = ['display', 'slide']
        for widget, name in zip(self.mc.displays['window'].walk(),
                                widget_hierarchy):
            getattr(self, 'check_{}'.format(name))(widget=widget)

    def check_display(self, widget):
        self.assertTrue(isinstance(widget, Display))

    def check_slide(self, widget):
        self.assertTrue(isinstance(widget, Slide))

    def test_current_slide_properties(self):
        # make sure we get the default blank slide
        self.assertEqual(self.mc.displays['window'].current_slide_name,
                         'window_blank')

        slide1 = Slide(mc=self.mc, name='slide1')
        self.mc.targets['default'].current_slide_name = 'slide1'

        # test display properties
        self.assertEqual(self.mc.displays['window'].current_slide, slide1)
        self.assertEqual(self.mc.displays['window'].current_slide_name,
                         'slide1')

        # test target properties
        self.assertEqual(
                self.mc.targets['window'].current_slide,
                slide1)
        self.assertEqual(
                self.mc.targets['window'].current_slide_name,
                'slide1')

        # make sure showing a slide at the same priority replaces the current
        # one.
        slide2 = Slide(mc=self.mc, name='slide2')
        self.mc.targets['window'].show_slide('slide2')

        self.assertEqual(self.mc.targets['window'].current_slide, slide2)
        self.assertEqual(self.mc.targets['window'].current_slide_name,
                         'slide2')
        self.assertEqual(self.mc.displays['window'].current_slide_name,
                         'slide2')
        self.assertEqual(self.mc.displays['window'].current_slide, slide2)

        # test property setters for displays
        self.mc.targets['window'].current_slide_name = 'slide1'
        self.assertEqual(self.mc.targets['window'].current_slide, slide1)

        self.mc.targets['window'].current_slide_name = 'slide2'
        self.assertEqual(self.mc.targets['window'].current_slide, slide2)

        self.mc.targets['window'].current_slide = slide1
        self.assertEqual(self.mc.targets['window'].current_slide, slide1)

        # we need to wait a tick to remove the slide we just added
        self.advance_time()

        # now remove the current slide and make sure slide1 comes back
        self.mc.targets['window'].remove_slide(slide2)
        self.assertEqual(self.mc.targets['window'].current_slide, slide1)

        # add another slide so we have 3
        slide3 = Slide(mc=self.mc, name='slide3', config={})
        self.assertEqual(len(self.mc.targets['window'].slides), 3)

        # also test removing by name
        self.mc.targets['window'].remove_slide('slide1')
        self.assertEqual(self.mc.targets['window'].current_slide, slide3)

    def test_priorities(self):
        # show slide 1, p100
        slide1 = Slide(mc=self.mc, name='slide1', priority=100)
        self.mc.targets['window'].show_slide('slide1')
        self.assertEqual(self.mc.targets['window'].current_slide, slide1)

        # show slide 2, p0, it should not show
        slide2 = Slide(mc=self.mc, name='slide2', priority=0)
        self.mc.targets['window'].show_slide('slide2')
        self.assertEqual(self.mc.targets['window'].current_slide, slide1)

        # show slide 3, p200, it should show
        slide3 = Slide(mc=self.mc, name='slide3', config={}, priority=200)
        self.mc.targets['window'].show_slide('slide3')
        self.assertEqual(self.mc.targets['window'].current_slide, slide3)

        # show slide 4, p199, it should not show
        slide4 = Slide(mc=self.mc, name='slide4', config={}, priority=199)
        self.mc.targets['window'].show_slide('slide4')
        self.assertLess(slide4.priority,
                        self.mc.targets['window'].current_slide.priority)

        # confirm that slide 3 is still current
        self.assertEqual(self.mc.targets['window'].current_slide, slide3)

        # force slide 4 to show
        self.assertLess(slide4.priority,
                        self.mc.targets['window'].current_slide.priority)
        self.mc.targets['window'].show_slide('slide4', force=True)

        # slide should show
        self.assertEqual(self.mc.targets['window'].current_slide_name,
                         'slide4')




"""Audio module provides all the audio features (playing of sounds) for the media controller."""
import logging

from kivy.clock import Clock
from mpf.core.case_insensitive_dict import CaseInsensitiveDict
from mpf.core.utility_functions import Util
from mpfmc.core.audio.audio_interface import AudioInterface
from mpfmc.core.audio.audio_exception import AudioException

__all__ = ('SoundSystem',
           'AudioInterface',
           'AudioException')

# ---------------------------------------------------------------------------
#    Default sound system and track values
# ---------------------------------------------------------------------------
DEFAULT_AUDIO_ENABLED = True
DEFAULT_AUDIO_BUFFER_SAMPLE_SIZE = 2048
DEFAULT_SAMPLE_RATE = 44100
DEFAULT_AUDIO_CHANNELS = 1
DEFAULT_MASTER_VOLUME = 0.5
DEFAULT_TRACK_MAX_SIMULTANEOUS_SOUNDS = 1
DEFAULT_TRACK_VOLUME = 0.5


# pylint: disable=too-many-instance-attributes
class SoundSystem:

    """Sound system for MPF.

    The SoundSystem class is used to read the sound system settings from the
    config file and then initialize the audio interface and create the
    specified tracks.
    """

    # pylint: disable=invalid-name, too-many-branches
    def __init__(self, mc):
        """Initialise sound system."""
        self.mc = mc
        self.log = logging.getLogger('SoundSystem')
        self._initialized = False
        self.audio_interface = None
        self.config = dict()
        self.sound_events = dict()
        self.tracks = CaseInsensitiveDict()
        self.clock_event = None

        self.log.debug("Loading the Sound System")

        # Load configuration for sound system
        if 'sound_system' not in self.mc.machine_config:
            self.log.info("SoundSystem: Using default 'sound_system' settings")
            self.config = dict()
        else:
            self.config = self.mc.machine_config['sound_system']

        # TODO: Use config spec validator

        # Validate configuration and provide default values where needed
        if 'enabled' not in self.config:
            self.config['enabled'] = DEFAULT_AUDIO_ENABLED

        # If the sound system has been disabled, abort initialization
        if not self.config['enabled']:
            self.log.debug("SoundSystem: The sound system has been disabled in "
                           "the configuration file (enabled: False). No audio "
                           "features will be available.")
            return

        if 'buffer' not in self.config or self.config['buffer'] == 'auto':
            self.config['buffer'] = DEFAULT_AUDIO_BUFFER_SAMPLE_SIZE
        elif not AudioInterface.power_of_two(self.config['buffer']):
            self.log.warning("SoundSystem: The buffer setting is not a power of "
                             "two. Default buffer size will be used.")
            self.config['buffer'] = DEFAULT_AUDIO_BUFFER_SAMPLE_SIZE

        if 'frequency' not in self.config or self.config['frequency'] == 'auto':
            self.config['frequency'] = DEFAULT_SAMPLE_RATE

        if 'channels' not in self.config:
            self.config['channels'] = DEFAULT_AUDIO_CHANNELS

        if 'master_volume' not in self.config:
            self.config['master_volume'] = DEFAULT_MASTER_VOLUME

        # Initialize audio interface library (get audio output)
        try:
            self.audio_interface = AudioInterface(
                rate=self.config['frequency'],
                channels=self.config['channels'],
                buffer_samples=self.config['buffer'])
        except AudioException:
            self.log.error("Could not initialize the audio interface. "
                           "Audio features will not be available.")
            self.audio_interface = None
            return

        # Setup tracks in audio system (including initial volume levels)
        if 'tracks' in self.config:
            for track_name, track_config in self.config['tracks'].items():
                self._create_track(track_name, track_config)
        else:
            self._create_track('default')
            self.log.info("No audio tracks are specified in your machine config file. "
                          "a track named 'default' has been created.")

        # Set initial master volume level
        self.master_volume = self.config['master_volume']

        # Establish machine tick function callback (will process internal audio events)
        self.clock_event = Clock.schedule_interval(self.tick, 0)

        # Start audio engine processing
        self.audio_interface.enable()
        self._initialized = True

        self.mc.events.add_handler("master_volume_increase", self.master_volume_increase)
        self.mc.events.add_handler("master_volume_decrease", self.master_volume_decrease)
        self.mc.events.add_handler("shutdown", self.shutdown)
        self.mc.events.add_handler("client_connected", self._send_volume, -1)

    def _send_volume(self, **kwargs):
        del kwargs
        self.mc.set_machine_var("master_volume", self.audio_interface.get_master_volume())

    def shutdown(self, **kwargs):
        """Shuts down the audio interface"""
        del kwargs
        if self.enabled:
            self.audio_interface.shutdown()
            self._initialized = False

    @property
    def enabled(self):
        """Return true if enabled."""
        return self._initialized

    @property
    def master_volume(self) -> float:
        """Return master volume."""
        return self.audio_interface.get_master_volume()

    @master_volume.setter
    def master_volume(self, value: float):
        """Set master volume."""
        # Constrain volume to the range 0.0 to 1.0
        value = min(max(value, 0.0), 1.0)
        self.audio_interface.set_master_volume(value)
        self._send_volume()

    @property
    def default_track(self):
        """Return default track."""
        return self.audio_interface.get_track(0)

    def master_volume_increase(self, delta: float = 0.05, **kwargs):
        """Increase master volume by delta.

        Args:
            delta: How much to increase volume?
        """
        del kwargs
        self.master_volume += delta
        self.log.info("Increased master volume by %s to %s.", delta, self.master_volume)

    def master_volume_decrease(self, delta: float = 0.05, **kwargs):
        """Decrease master volume by delta.

        Args:
            delta: How much to decrease volume?
        """
        del kwargs
        self.master_volume -= delta
        self.log.info("Decreased master volume by %s to %s.", delta, self.master_volume)

    def _create_track(self, name, config=None):     # noqa
        """Create a track in the audio system with the specified name and configuration.

        Args:
            name: The track name (which will be used to reference the track, such as
                "voice" or "sfx".
            config: A Python dictionary containing the configuration settings for
                this track.
        """
        if self.audio_interface is None:
            raise AudioException("Could not create '{}' track - the sound_system has "
                                 "not been initialized".format(name))

        # Validate track config parameters
        if name in self.tracks:
            raise AudioException("Could not create '{}' track - a track with that name "
                                 "already exists".format(name))

        if config is None:
            config = {}

        if 'volume' not in config:
            config['volume'] = DEFAULT_TRACK_VOLUME

        if 'type' not in config:
            config['type'] = 'standard'

        if config['type'] not in ['standard', 'playlist', 'sound_loop']:
            raise AudioException("Could not create '{}' track - an illegal value for "
                                 "'type' was found".format(name))

        # Validate type-specific parameters and create the track
        track = None
        if config['type'] == 'standard':
            if 'simultaneous_sounds' not in config:
                config['simultaneous_sounds'] = DEFAULT_TRACK_MAX_SIMULTANEOUS_SOUNDS

            track = self.audio_interface.create_standard_track(self.mc,
                                                               name,
                                                               config['simultaneous_sounds'],
                                                               config['volume'])
        elif config['type'] == 'playlist':
            config.setdefault('crossfade_time', 0.0)
            config['crossfade_time'] = Util.string_to_secs(config['crossfade_time'])

            track = self.audio_interface.create_playlist_track(self.mc,
                                                               name,
                                                               config['crossfade_time'],
                                                               config['volume'])

        elif config['type'] == 'sound_loop':
            if 'max_layers' not in config:
                config['max_layers'] = 8

            track = self.audio_interface.create_sound_loop_track(self.mc,
                                                                 name,
                                                                 config['max_layers'],
                                                                 config['volume'])

        if track is None:
            raise AudioException("Could not create '{}' track due to an error".format(name))

        self.tracks[name] = track

        if 'events_when_stopped' in config and config['events_when_stopped'] is not None:
            track.events_when_stopped = Util.string_to_event_list(config['events_when_stopped'])

        if 'events_when_played' in config and config['events_when_played'] is not None:
            track.events_when_played = Util.string_to_event_list(config['events_when_played'])

        if 'events_when_paused' in config and config['events_when_paused'] is not None:
            track.events_when_paused = Util.string_to_event_list(config['events_when_paused'])

        if 'events_when_resumed' in config and config['events_when_resumed'] is not None:
            track.events_when_resumed = Util.string_to_event_list(config['events_when_resumed'])

    def tick(self, dt):
        """Clock callback function"""
        del dt
        self.audio_interface.process()



from mpfmc.tests.MpfIntegrationTestCase import MpfIntegrationTestCase

try:
    from mpfmc.core.audio import SoundSystem
except ImportError:
    SoundSystem = None


class TestAudio(MpfIntegrationTestCase):

    def get_machine_path(self):
        return 'integration/machine_files/audio'

    def get_config_file(self):
        return 'config.yaml'

    def test_sound_custom_code(self):

        if SoundSystem is None or self.mc.sound_system is None:
            self.skipTest("Sound system is not enabled")

        self.assertIsNotNone(self.mc.sound_system)
        interface = self.mc.sound_system.audio_interface
        if interface is None:
            self.skipTest("Sound system audio interface could not be loaded")

        self.assertIsNotNone(interface)

        # Check sfx track
        track_sfx = interface.get_track_by_name("sfx")
        self.assertIsNotNone(track_sfx)
        self.assertEqual(track_sfx.name, "sfx")
        self.assertAlmostEqual(track_sfx.volume, 0.6, 1)

        # /sounds/sfx
        self.assertTrue(hasattr(self.mc, 'sounds'))
        self.assertIn('264828_text', self.mc.sounds)

        # Attempt to play sound by calling sound_player directly from custom code
        settings = {
            '264828_text': {
                'action': 'play',
                'loops': -1,
                'key': 'music'
            }
        }
        self.machine.sound_player.play(settings, 'asset_manager', None)
        self.advance_time_and_run(0.2)

        status = track_sfx.get_status()
        self.assertEqual(status[0]['status'], "playing")
        self.assertEqual(status[0]['sound_id'], self.mc.sounds["264828_text"].id)
        self.assertEqual(len(status), 1)

        settings = {
            '264828_text': {
                'action': 'stop'
            }
        }
        self.machine.sound_player.play(settings, 'asset_manager', None)
        self.advance_time_and_run(0.1)

    def test_sound_loop_player(self):

        if SoundSystem is None or self.mc.sound_system is None:
            self.skipTest("Sound system is not enabled")

        self.assertIsNotNone(self.mc.sound_system)
        interface = self.mc.sound_system.audio_interface
        if interface is None:
            self.skipTest("Sound system audio interface could not be loaded")

        self.assertIsNotNone(interface)

        # start mode1, its slide with video should come up
        self.post_event("start_mode1")
        self.advance_time_and_run()

        self.assertEqual(self.mc.targets['default'].current_slide.name, 'mode1_slide1')

        # Check sound loop track
        track_loops = interface.get_track_by_name("loops")
        self.assertIsNotNone(track_loops)
        self.assertEqual(track_loops.name, "loops")
        self.assertAlmostEqual(track_loops.volume, 0.6, 1)

        # /sounds/loops
        self.assertTrue(hasattr(self.mc, 'sounds'))
        self.assertIn('kick', self.mc.sounds)

        # Sound loop sets
        self.assertTrue(hasattr(self.mc, 'sound_loop_sets'))
        self.assertIn('basic_beat', self.mc.sound_loop_sets)

        # Test sound_loop_player
        self.advance_time_and_run()
        self.post_event('play_basic_beat')
        self.advance_time_and_run(6)

        status = track_loops.get_status()
        self.assertEqual(status[0]['status'], "playing")
        self.assertEqual(status[0]['sound_id'], self.mc.sounds["kick"].id)
        self.assertEqual(status[0]['stop_loop_samples_remaining'], "DO NOT STOP LOOP")
        self.assertEqual(len(status), 1)

        self.post_event('stop_current_loop')
        self.advance_time_and_run(3)

    def test_sound_player_and_show(self):

        if SoundSystem is None or self.mc.sound_system is None:
            self.skipTest("Sound system is not enabled")

        self.assertIsNotNone(self.mc.sound_system)
        interface = self.mc.sound_system.audio_interface
        if interface is None:
            self.skipTest("Sound system audio interface could not be loaded")

        self.assertIsNotNone(interface)

        # Make sure sound assets are present
        self.assertTrue(hasattr(self.mc, 'sounds'))
        self.assertIn('264828_text', self.mc.sounds)

        self.mock_event("use_sound_setting")
        self.mock_event("text_sound_played")
        self.mock_event("text_sound_played_from_show")
        self.mock_event("text_sound_played_from_sound_player")
        self.mock_event("text_sound_stopped")
        self.mock_event("text_sound_stopped_from_show")
        self.mock_event("text_sound_stopped_from_sound_player")

        # Play sound using sound player
        self.post_event('play_sound_1')
        self.advance_time_and_run(0.1)
        self.assertEventNotCalled("use_sound_setting")
        self.assertEventNotCalled("text_sound_played")
        self.assertEventCalled("text_sound_played_from_sound_player")

        self.post_event("stop_sound")
        self.advance_time_and_run(0.1)
        self.assertEventNotCalled("use_sound_setting")
        self.assertEventNotCalled("text_sound_stopped")
        self.assertEventCalled("text_sound_stopped_from_sound_player")
        self.reset_mock_events()

        self.post_event('play_sound_2')
        self.advance_time_and_run(0.1)
        self.assertEventNotCalled("use_sound_setting")
        self.assertEventNotCalled("text_sound_played_from_sound_player")
        self.assertEventCalled("text_sound_played")

        self.post_event("stop_sound")
        self.advance_time_and_run(0.1)
        self.assertEventNotCalled("use_sound_setting")
        self.assertEventCalled("text_sound_stopped")
        self.assertEventNotCalled("text_sound_stopped_from_sound_player")
        self.reset_mock_events()

        self.post_event('play_sound_3')
        self.advance_time_and_run(0.1)
        self.assertEventNotCalled("use_sound_setting")
        self.assertEventNotCalled("text_sound_played_from_sound_player")
        self.assertEventNotCalled("text_sound_played")

        self.post_event("stop_sound")
        self.advance_time_and_run(0.1)
        self.assertEventNotCalled("use_sound_setting")
        self.assertEventCalled("text_sound_stopped")
        self.assertEventNotCalled("text_sound_stopped_from_sound_player")
        self.reset_mock_events()

        # Play first show
        self.post_event('play_sound_test_1_show')
        self.advance_time_and_run(0.1)
        self.assertEventNotCalled("use_sound_setting")
        self.assertEventNotCalled("text_sound_played")
        self.assertEventCalled("text_sound_played_from_show")

        self.post_event("stop_sound")
        self.advance_time_and_run(0.1)
        self.assertEventNotCalled("use_sound_setting")
        self.assertEventNotCalled("text_sound_stopped")
        self.assertEventCalled("text_sound_stopped_from_show")
        self.reset_mock_events()

        # Play second show
        self.post_event('play_sound_test_2_show')
        self.advance_time_and_run(0.1)
        self.assertEventNotCalled("use_sound_setting")
        self.assertEventNotCalled("text_sound_played_from_show")
        self.assertEventCalled("text_sound_played")

        self.post_event("stop_sound")
        self.advance_time_and_run(0.1)
        self.assertEventNotCalled("use_sound_setting")
        self.assertEventCalled("text_sound_stopped")
        self.assertEventNotCalled("text_sound_stopped_from_show")
        self.reset_mock_events()

        # Play third show
        self.post_event('play_sound_test_3_show')
        self.advance_time_and_run(0.1)
        self.assertEventNotCalled("use_sound_setting")
        self.assertEventNotCalled("text_sound_played_from_show")
        self.assertEventNotCalled("text_sound_played")

        self.post_event("stop_sound")
        self.advance_time_and_run(0.1)
        self.assertEventNotCalled("use_sound_setting")
        self.assertEventCalled("text_sound_stopped")
        self.assertEventNotCalled("text_sound_stopped_from_show")



from mpfmc.tests.MpfMcTestCase import MpfMcTestCase


class TestWidgetStyles(MpfMcTestCase):
    def get_machine_path(self):
        return 'tests/machine_files/widget_styles'

    def get_config_file(self):
        return 'test_widget_styles.yaml'

    def get_widget(self, index=0):
        return self.mc.targets['default'].current_slide.widgets[index].widget

    def test_style_loading(self):
        self.assertIn('text_default', self.mc.machine_config['widget_styles'])

    def test_style(self):
        self.mc.events.post('slide1')
        self.advance_time()

        # Test named style

        # font size set in style
        self.assertEqual(self.get_widget().font_size, 100)

        # halign set in style and widget, widget should win
        self.assertEqual(self.get_widget().halign, 'right')

        # second widget has no style set, so it should get the default
        self.assertEqual(self.get_widget(1).font_size, 21)

    def test_default_style(self):
        self.mc.events.post('slide3')
        self.advance_time()

        self.assertEqual(self.get_widget().color, [1, 0, 0, 1])

    def test_invalid_style(self):
        self.mc.events.post('slide4')

        with self.assertRaises(Exception) as e:
            self.advance_time()

        self.assertIsInstance(e.exception.__cause__, ValueError)

    def test_local_setting_overrides_style(self):
        self.mc.events.post('slide5')
        self.advance_time()

        self.assertEqual(self.get_widget().font_size, 50)

    def test_stacked_styles(self):
        self.mc.events.post('slide6')
        self.advance_time()
        self.assertEqual(self.get_widget().font_size, 100)
        self.assertEqual(self.get_widget().color, [0.0, 0.0, 1.0, 1]);

    def test_stacked_order_overrides(self):
        self.mc.events.post('slide7')
        self.advance_time()
        self.assertEqual(self.get_widget().font_size, 21)
        self.assertEqual(self.get_widget().color, [0.0, 0.0, 1.0, 1]);

    # todo some future release

    # def test_mode_style(self):
    #     self.mc.modes['mode1'].start()
    #     self.advance_time()
    #
    #     self.mc.events.post('slide2')
    #     self.advance_time()
    #
    #     # widget with no style, should pickup default out of the mode
    #     # text_strings, rather than the machine wide one
    #     self.assertEqual(self.get_widget().font_size, 50)
    #
    #     # mode widget with style from machine wide config
    #     self.assertEqual(self.get_widget(1).font_size, 100)
    #
    #     # mode widget with style name that's not valid, so it should
    #     # pickup the default
    #     self.assertEqual(self.get_widget(2).font_size, 50)



from mpfmc.tests.MpfMcTestCase import MpfMcTestCase
from mpfmc.widgets.video import VideoWidget


class TestVideo(MpfMcTestCase):
    def get_machine_path(self):
        return 'tests/machine_files/video'

    def get_config_file(self):
        return 'test_video.yaml'

    def test_video(self):

        # Note the green bar on the bottom of this video is part of it. It is
        # VERY low quality to keep the mpf-mc package small. :)
        self.assertIn('mpf_video_small_test', self.mc.videos)

        self.mc.events.post('show_slide1')
        self.advance_time()

        video_widget = self.mc.targets['default'].current_slide.widgets[0]
        text = self.mc.targets['default'].current_slide.widgets[2]

        self.assertEqual(video_widget.state, 'play')
        self.assertTrue(video_widget.video.loaded)

        # removed since this is a race condition, pos = -1 if the video hasn't
        # actually started playing yet.
        # self.assertAlmostEqual(video_widget.position, 0.0, delta=.3)

        self.assertAlmostEqual(video_widget.video.duration, 7.96, delta=.5)
        self.assertEqual(video_widget.video.volume, 1.0)

        text.text = 'PLAY'
        self.advance_time(1)

        # now that 1 sec has passed, make sure the video is advancing
        self.assertGreater(video_widget.position, 0)

        # also check the size. The size isn't set until the video actually
        # starts playing, which is why we don't check it until now.
        self.assertEqual(video_widget.size, [100, 70])

        video_widget.pause()
        text.text = 'PAUSE'
        self.advance_time(.1)

        self.assertEqual('paused', video_widget.video.state)
        pos = video_widget.video.position
        self.advance_time(.5)

        # make sure it's really stopped
        self.assertEqual(pos, video_widget.video.position)

        # now start it again
        video_widget.play()
        text.text = 'PLAY'
        self.advance_time(1)

        self.assertEqual('playing', video_widget.video.state)
        self.assertGreater(video_widget.video.position, pos)

        # stop it, should go back to the beginning
        video_widget.stop()
        text.text = 'STOP'
        self.advance_time(1)
        self.assertEqual(0.0, video_widget.video.position)

        video_widget.play()
        text.text = 'PLAY FROM BEGINNING'
        self.advance_time(1)
        self.assertAlmostEqual(video_widget.video.position, 1.0, delta=.5)

        # jump to the 50% point
        video_widget.seek(.5)
        text.text = 'SEEK TO 50%'
        self.advance_time(.1)
        self.assertAlmostEqual(video_widget.video.position, 4.0, delta=.5)

        # test the volume
        video_widget.set_volume(.5)
        text.text = 'VOLUME 50% (though this vid has no audible sound)'
        self.assertEqual(.5, video_widget.video.volume)

        video_widget.set_volume(1)
        text.text = 'VOLUME 100% (though this vid has no audible sound)'
        self.assertEqual(1.0, video_widget.video.volume)

        video_widget.set_position(.5)
        text.text = 'SET POSITION TO 0.5s'
        self.advance_time(.1)
        self.assertAlmostEqual(video_widget.video.position, 0.6, delta=.5)

        # jump to the 90% point
        video_widget.seek(.9)
        video_widget.play()
        video_widget.video.set_end_behavior('stop')
        text.text = 'JUMP TO 90%, STOP AT END'
        self.advance_time(2)
        self.assertEqual(video_widget.video.position, 0)

        # jump to the 90% point
        video_widget.seek(.9)
        video_widget.play()
        video_widget.video.set_end_behavior('pause')
        text.text = 'JUMP TO 90%, PAUSE AT END'
        self.advance_time(2)
        self.assertAlmostEqual(video_widget.video.position, 7.9, delta=.5)

        # jump to the 90% point
        video_widget.seek(.9)
        video_widget.play()
        video_widget.video.set_end_behavior('loop')
        text.text = 'JUMP TO 90%, LOOP AT END'
        self.advance_time(2)
        self.assertAlmostEqual(video_widget.video.position, 1, delta=.5)

    def test_control_events(self):
        self.mc.events.post('show_slide2')
        self.advance_real_time(1)

        video_widget = self.mc.targets['default'].current_slide.widgets[0].widget
        text = self.mc.targets['default'].current_slide.widgets[2].widget
        text.text = "PLAY"

        self.assertEqual('playing', video_widget.video.state)
        self.assertGreater(video_widget.video.position, 0)

        self.mc.events.post('stop1')
        text.text = 'STOP'
        self.advance_real_time()
        self.assertEqual(0.0, video_widget.video.position)

        self.mc.events.post('play1')
        text.text = 'PLAY'
        self.advance_real_time(1)
        self.assertGreater(video_widget.video.position, 0)

        self.mc.events.post('pause1')
        text.text = 'PAUSE'
        self.advance_real_time()
        pos = video_widget.video.position
        self.advance_real_time()
        self.assertEqual(video_widget.video.position, pos)

        self.mc.events.post('seek1')
        text.text = 'SEEK'
        self.advance_real_time()
        self.assertAlmostEqual(video_widget.video.position, 3.5, delta=.5)

        self.mc.events.post('mute')
        text.text = 'VOLUME (MUTE)'
        self.advance_real_time()
        self.assertEqual(0, video_widget.video.volume)

        self.mc.events.post('position1')
        text.text = "JUMP TO POSITION 4s"
        self.advance_real_time()
        self.assertAlmostEqual(video_widget.video.position, 4.0, delta=.1)

        # remove widget and verify control events are removed
        self.mc.targets['default'].remove_slide('video_test2')
        self.advance_time()
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'default_blank')

        self.assertFalse(self.mc.events.does_event_exist('play1'))
        self.assertFalse(self.mc.events.does_event_exist('stop1'))
        self.assertFalse(self.mc.events.does_event_exist('pause1'))
        self.assertFalse(self.mc.events.does_event_exist('seek1'))
        self.assertFalse(self.mc.events.does_event_exist('mute'))
        self.assertFalse(self.mc.events.does_event_exist('position1'))

    def test_video_settings(self):
        self.mc.events.post('show_slide7')
        self.advance_time(1)
        video_widget = self.mc.targets['default'].current_slide.widgets[0]

        # make sure it's playing
        self.assertEqual('playing', video_widget.video.state)
        self.assertEqual(0.2, video_widget.video.volume)
        last_pos = video_widget.video.position
        self.advance_time(1)

        # make sure the current position is about 1 sec later than last
        self.assertAlmostEqual(1, video_widget.video.position - last_pos, delta=.2)

        # jump to 90%
        self.mc.events.post('seek1')
        self.advance_time()
        self.assertGreater(video_widget.video.position, 7)

        # wait for the video to end and make sure it loops
        self.advance_time(2)
        last_pos = video_widget.video.position
        self.advance_time(1)

        # should still be playing, make sure the current position is about 1
        # sec later than last
        self.assertAlmostEqual(1, video_widget.video.position - last_pos, delta=.2)

        # stop and reset the video since it's the same asset used in the next
        # slide
        video_widget.stop()
        self.advance_time(.1)

        self.mc.events.post('show_slide8')
        self.advance_time(1)
        video_widget = self.mc.targets['default'].current_slide.widgets[0]

        self.assertEqual('', video_widget.video.state)
        # stopped state is empty string

        self.assertEqual(video_widget.video.position, 0.0)
        self.advance_time()
        self.assertEqual(video_widget.video.position, 0.0)
        self.assertEqual(0.8, video_widget.video.volume)

        self.mc.events.post('play1')
        self.advance_time()

        # wait for the video to end and make sure it stops
        self.mc.events.post('seek1')
        self.advance_time(2)
        self.assertEqual(video_widget.video.position, 0.0)

        self.advance_time(1)
        self.assertEqual(video_widget.video.position, 0.0)

    def test_pre_show_slide(self):
        self.mc.events.post('show_slide3')
        self.advance_time()

    def test_show_slide(self):
        self.mc.events.post('show_slide4')
        self.advance_time()

    def test_pre_slide_leave(self):
        self.mc.events.post('show_slide5')
        self.advance_time()

    def test_slide_leave(self):
        self.mc.events.post('show_slide6')
        self.advance_time()

    def test_video_stops_on_slide_removal(self):
        self.mc.events.post('show_slide9')
        self.advance_time()
        self.assertEqual(self.mc.targets['default'].current_slide.name, 'video_test9')

        # start mode1, its slide with video should come up
        self.mc.modes['mode1'].start()
        self.advance_time()

        self.mc.events.post('mode_mode1_started')
        self.advance_time(2)
        self.assertEqual(self.mc.targets['default'].current_slide.name, 'mode1_slide1')

        video_widget = self.mc.targets['default'].current_slide.widgets[0]
        self.assertTrue(isinstance(video_widget, VideoWidget))
        self.assertEqual(video_widget.state, 'play')

        # stop mode 1, should unload the slide
        self.mc.modes['mode1'].stop()
        self.advance_time()
        self.assertEqual(self.mc.targets['default'].current_slide.name, 'video_test9')

        self.assertEqual(video_widget.state, 'stop')



"""Contains the sound config player class."""
# WARNING: Do not import kivy's logger here since that will trigger Kivy to
# load in the mpf process when MPF processes the MpfSoundPlayer
from copy import deepcopy

from mpfmc.core.mc_config_player import McConfigPlayer


class McTrackPlayer(McConfigPlayer):

    """Base class for the Track Player that runs on the mpf-mc side of things.

    It receives all of its instructions via BCP from a MpfTrackPlayer instance
    running as part of MPF.

    The track_player: section of a config file (either the machine-wide or
    a mode-specific config) looks like this:

    track_player:
        <event_name>:
            <track_name>:
                <track_settings>: ...

    There is no express config for the track_player.

    To control the track settings (such as volume and action), enter the track name
    on the next line and the settings indented under it, like this:

    track_player:
        some_event:
            track_name:
                action: pause
                fade: 0.5 sec

    Here are several various examples:

    track_player:
        some_event:
            music:
                action: set_volume
                volume: 0.65
                fade: 2 sec

        some_event2:
            sfx:
                action: stop
                fade: 0.75 sec

    """

    config_file_section = 'track_player'
    show_section = 'tracks'
    machine_collection_name = None

    # pylint: disable=invalid-name
    def play(self, settings, context, calling_context, priority=0, **kwargs):
        """Initiates an action from a validated tracks: section from a track_player: section of a
        config file or the tracks: section of a show.

        The config must be validated. Validated config looks like this:

        <track_name>:
            <settings>: ...

        <settings> can be:

        action:
        volume:
        fade:
        """
        del priority
        del calling_context
        settings = deepcopy(settings)

        if 'tracks' in settings:
            settings = settings['tracks']

        for track_name, s in settings.items():

            selected_tracks = dict()

            # Track name '__all__' indicates the action should be performed on all tracks
            if track_name == "__all__":
                selected_tracks = self.machine.sound_system.tracks
            else:
                # Single track name specified - retrieve track by name
                try:
                    selected_track = self.machine.sound_system.tracks[track_name]
                    selected_tracks[selected_track.name] = selected_track
                except KeyError:
                    self.machine.log.error("TrackPlayer: The specified audio track "
                                           "does not exist ('{}').".format(track_name))
                    return

            s.update(kwargs)

            # TODO: perform validation on parameters

            # Loop over selected tracks performing action with settings on each one
            for track in selected_tracks.values():

                # Determine action to perform
                if s['action'].lower() == 'play':
                    track.play(s['fade'])

                elif s['action'].lower() == 'stop':
                    track.stop(s['fade'])

                elif s['action'].lower() == 'pause':
                    track.pause(s['fade'])

                elif s['action'].lower() == 'set_volume':
                    track.set_volume(s['volume'], s['fade'])

                elif s['action'].lower() == 'stop_all_sounds':
                    track.stop_all_sounds(s['fade'])

                else:
                    self.machine.log.error("TrackPlayer: The specified action "
                                           "is not valid ('%s').", s['action'])

    def get_express_config(self, value):
        """ express config for tracks is not supported"""
        del value
        raise AssertionError("Track Player does not support express config")

    # pylint: disable=too-many-branches
    def validate_config(self, config):
        """Validates the track_player: section of a config file (either a
        machine-wide config or a mode config).

        Args:
            config: A dict of the contents of the track_player section
            from the config file. It's assumed that keys are event names, and
            values are settings for what the track_player should do when that
            event is posted.

        Returns: A dict a validated entries.

        This method overrides the base method since the track_player has
        unique options.

        """
        # first, we're looking to see if we have a string, a list, or a dict.
        # if it's a dict, we look to see whether we have the name of some sound

        validated_config = dict()

        for event, settings in config.items():
            validated_config[event] = dict()
            validated_config[event]['tracks'] = dict()

            if not isinstance(settings, dict):
                settings = {settings: dict()}

            for track, track_settings in settings.items():

                # Now check to see if all the settings are valid
                # sound settings. If not, assume it's a single sound settings.
                if isinstance(track_settings, dict):
                    for key in track_settings:
                        if key not in self.machine.config_validator.get_config_spec()['track_player']:
                            break

                    validated_config[event]['tracks'].update(
                        self._validate_config_item(track, track_settings))

        return validated_config

    def _validate_config_item(self, device, device_settings):
        """Validates the config when in a show"""
        validated_dict = super()._validate_config_item(device, device_settings)
        # device is track name

        # Ensure volume parameter value has been provided for 'set_volume' actions
        if validated_dict[device]['action'] == 'set_volume' and \
                validated_dict[device]['volume'] is None:
            raise ValueError("track_player: 'volume' must be provided for all 'set_volume' "
                             "actions ({} track)".format(device))

        if validated_dict[device]['fade'] < 0:
            raise ValueError("track_player: 'fade' must be greater than or equal to zero for all "
                             "actions ({} track)".format(device))

        return validated_dict

    def clear_context(self, context):
        """Because tracks are persistent for the life of the application, there is nothing
        to clear when the context ends. No new track instances are created using the
        track_player."""
        del context


mc_player_cls = McTrackPlayer



from os import path

from mpfmc.uix.bitmap_font.bitmap_font import BitmapFont

from mpfmc.assets.mc_asset import McAsset


class BitmapFontAsset(McAsset):

    attribute = 'bitmap_fonts'
    path_string = 'bitmap_fonts'
    config_section = 'bitmap_fonts'
    extensions = ('png', 'gif', 'bmp')
    class_priority = 100
    pool_config_section = None
    asset_group_class = None

    def __init__(self, mc, name, file, config):
        super().__init__(mc, name, file, config)

        self._bitmap_font = None  # holds the actual image and font info in memory

        # Validate the descriptor setting (it can contain either a list, or a
        # descriptor file name).  If the descriptor setting is omitted, a file
        # will be used with the same name as the font image asset file, but with a
        # .fnt extension.
        if 'descriptor' not in self.config or not self.config['descriptor']:
            self.config['descriptor'] = path.splitext(self.config['file'])[0] + '.fnt'

        if isinstance(self.config['descriptor'], str):
            if not path.isfile(self.config['descriptor']):
                raise FileNotFoundError('Could not locate the bitmap font descriptor file {}'.format(
                                        self.config['descriptor']))

        elif not isinstance(self.config['descriptor'], list):
            raise ValueError('Bitmap font descriptor must contain either a list or file name.')

    @property
    def bitmap_font(self):
        return self._bitmap_font

    def get_extents(self, text, font_kerning=True):
        if self._bitmap_font:
            return self.bitmap_font.get_extents(text, font_kerning)

        return 0, 0

    def get_descent(self):
        if self._bitmap_font:
            return self.bitmap_font.get_descent()

        return 0

    def get_ascent(self):
        if self._bitmap_font:
            return self.bitmap_font.get_ascent()

        return 0

    def do_load(self):
        # Load the bitmap font image atlas
        self._bitmap_font = BitmapFont(self.config['file'], self.config['descriptor'])

    def _do_unload(self):
        self._bitmap_font = None



"""Test widgets."""
import weakref

import gc

from mpfmc.uix.widget import WidgetContainer, Widget
from mpfmc.widgets.rectangle import Rectangle
from mpfmc.widgets.text import Text
from mpfmc.widgets.bezier import Bezier
from mpfmc.widgets.line import Line
from mpfmc.widgets.ellipse import Ellipse
from mpfmc.widgets.quad import Quad
from mpfmc.widgets.point import Point
from mpfmc.widgets.triangle import Triangle
from mpfmc.tests.MpfMcTestCase import MpfMcTestCase
from unittest.mock import MagicMock


class TestWidgetWithNamedColor(MpfMcTestCase):
    def get_machine_path(self):
        return 'tests/machine_files/widgets'

    def get_config_file(self):
        return 'test_widgets_with_named_colors.yaml'

    def test_named_colors(self):
        self.mc.events.post("add_widget1_to_current")
        self.advance_real_time()

class TestWidget(MpfMcTestCase):
    def get_machine_path(self):
        return 'tests/machine_files/widgets'

    def get_config_file(self):
        return 'test_widgets.yaml'

    def test_anchor_offset_position(self):
        # For all these tests, the widget is 10x10

        # No anchor set, widget should be centered in the parent. Parent is
        # 100x100, widget is 10x10, so center of the parent is 50, 50, and
        # lower left corner of the widget is 45, 45

        # test with all defaults
        config = {"width": 10, "height": 10, "type": "rectangle"}
        self.mc.config_validator.validate_config('widgets:rectangle', config, base_spec='widgets:common')

        widget = Rectangle(self.mc, config)
        self.assertEqual(widget.size, [10, 10])
        self.assertEqual(widget.anchor_offset_pos, (-5, -5))

        # test anchors

        # bottom left
        config = {"anchor_x": "left", "anchor_y": "bottom",
                  "width": 10, "height": 10, "type": "rectangle"}
        self.mc.config_validator.validate_config('widgets:rectangle', config, base_spec='widgets:common')

        widget = Rectangle(self.mc, config)
        self.assertEqual(widget.anchor_offset_pos, (0, 0))

        # add adjustments
        config = {"anchor_x": "left", "anchor_y": "bottom", "adjust_top": 2,
                  "width": 10, "height": 10, "type": "rectangle"}
        self.mc.config_validator.validate_config('widgets:rectangle', config, base_spec='widgets:common')
        widget = Rectangle(self.mc, config)
        self.assertEqual(widget.anchor_offset_pos, (0, 0))

        config = {"anchor_x": "left", "anchor_y": "bottom", "adjust_right": 2,
                  "width": 10, "height": 10, "type": "rectangle"}
        self.mc.config_validator.validate_config('widgets:rectangle', config, base_spec='widgets:common')
        widget = Rectangle(self.mc, config)
        self.assertEqual(widget.anchor_offset_pos, (0, 0))

        config = {"anchor_x": "left", "anchor_y": "bottom", "adjust_bottom": 2,
                  "width": 10, "height": 10, "type": "rectangle"}
        self.mc.config_validator.validate_config('widgets:rectangle', config, base_spec='widgets:common')
        widget = Rectangle(self.mc, config)
        self.assertEqual(widget.anchor_offset_pos, (0, -2))

        config = {"anchor_x": "left", "anchor_y": "bottom", "adjust_left": 2,
                  "width": 10, "height": 10, "type": "rectangle"}
        self.mc.config_validator.validate_config('widgets:rectangle', config, base_spec='widgets:common')
        widget = Rectangle(self.mc, config)
        self.assertEqual(widget.anchor_offset_pos, (-2, 0))

        # middle middle
        config = {"anchor_x": "middle", "anchor_y": "middle",
                  "width": 10, "height": 10, "type": "rectangle"}
        self.mc.config_validator.validate_config('widgets:rectangle', config, base_spec='widgets:common')
        widget = Rectangle(self.mc, config)
        self.assertEqual(widget.anchor_offset_pos, (-5, -5))

        # add adjustments
        config = {"anchor_x": "middle", "anchor_y": "middle", "adjust_top": 2,
                  "width": 10, "height": 10, "type": "rectangle"}
        self.mc.config_validator.validate_config('widgets:rectangle', config, base_spec='widgets:common')
        widget = Rectangle(self.mc, config)
        self.assertEqual(widget.anchor_offset_pos, (-5, -4))

        config = {"anchor_x": "middle", "anchor_y": "middle", "adjust_right": 2,
                  "width": 10, "height": 10, "type": "rectangle"}
        self.mc.config_validator.validate_config('widgets:rectangle', config, base_spec='widgets:common')
        widget = Rectangle(self.mc, config)
        self.assertEqual(widget.anchor_offset_pos, (-4, -5))

        config = {"anchor_x": "middle", "anchor_y": "middle", "adjust_bottom": 2,
                  "width": 10, "height": 10, "type": "rectangle"}
        self.mc.config_validator.validate_config('widgets:rectangle', config, base_spec='widgets:common')
        widget = Rectangle(self.mc, config)
        self.assertEqual(widget.anchor_offset_pos, (-5, -6))

        config = {"anchor_x": "middle", "anchor_y": "middle", "adjust_left": 2,
                  "width": 10, "height": 10, "type": "rectangle"}
        self.mc.config_validator.validate_config('widgets:rectangle', config, base_spec='widgets:common')
        widget = Rectangle(self.mc, config)
        self.assertEqual(widget.anchor_offset_pos, (-6, -5))

        # center center
        config = {"anchor_x": "center", "anchor_y": "center",
                  "width": 10, "height": 10, "type": "rectangle"}
        self.mc.config_validator.validate_config('widgets:rectangle', config, base_spec='widgets:common')
        widget = Rectangle(self.mc, config)
        self.assertEqual(widget.anchor_offset_pos, (-5, -5))

        # add adjustments
        config = {"anchor_x": "center", "anchor_y": "center", "adjust_top": 2,
                  "width": 10, "height": 10, "type": "rectangle"}
        self.mc.config_validator.validate_config('widgets:rectangle', config, base_spec='widgets:common')
        widget = Rectangle(self.mc, config)
        self.assertEqual(widget.anchor_offset_pos, (-5, -4))

        config = {"anchor_x": "center", "anchor_y": "center", "adjust_right": 2,
                  "width": 10, "height": 10, "type": "rectangle"}
        self.mc.config_validator.validate_config('widgets:rectangle', config, base_spec='widgets:common')
        widget = Rectangle(self.mc, config)
        self.assertEqual(widget.anchor_offset_pos, (-4, -5))

        config = {"anchor_x": "center", "anchor_y": "center", "adjust_bottom": 2,
                  "width": 10, "height": 10, "type": "rectangle"}
        self.mc.config_validator.validate_config('widgets:rectangle', config, base_spec='widgets:common')
        widget = Rectangle(self.mc, config)
        self.assertEqual(widget.anchor_offset_pos, (-5, -6))

        config = {"anchor_x": "center", "anchor_y": "center", "adjust_left": 2,
                  "width": 10, "height": 10, "type": "rectangle"}
        self.mc.config_validator.validate_config('widgets:rectangle', config, base_spec='widgets:common')
        widget = Rectangle(self.mc, config)
        self.assertEqual(widget.anchor_offset_pos, (-6, -5))

        # top right
        config = {"anchor_x": "right", "anchor_y": "top",
                  "width": 10, "height": 10, "type": "rectangle"}
        self.mc.config_validator.validate_config('widgets:rectangle', config, base_spec='widgets:common')

        widget = Rectangle(self.mc, config)
        self.assertEqual(widget.anchor_offset_pos, (-10, -10))

        # add adjustments
        config = {"anchor_x": "right", "anchor_y": "top", "adjust_top": 2,
                  "width": 10, "height": 10, "type": "rectangle"}
        self.mc.config_validator.validate_config('widgets:rectangle', config, base_spec='widgets:common')
        widget = Rectangle(self.mc, config)
        self.assertEqual(widget.anchor_offset_pos, (-10, -8))

        config = {"anchor_x": "right", "anchor_y": "top", "adjust_right": 2,
                  "width": 10, "height": 10, "type": "rectangle"}
        self.mc.config_validator.validate_config('widgets:rectangle', config, base_spec='widgets:common')
        widget = Rectangle(self.mc, config)
        self.assertEqual(widget.anchor_offset_pos, (-8, -10))

        config = {"anchor_x": "right", "anchor_y": "top", "adjust_bottom": 2,
                  "width": 10, "height": 10, "type": "rectangle"}
        self.mc.config_validator.validate_config('widgets:rectangle', config, base_spec='widgets:common')
        widget = Rectangle(self.mc, config)
        self.assertEqual(widget.anchor_offset_pos, (-10, -10))

        config = {"anchor_x": "right", "anchor_y": "top", "adjust_left": 2,
                  "width": 10, "height": 10, "type": "rectangle"}
        self.mc.config_validator.validate_config('widgets:rectangle', config, base_spec='widgets:common')
        widget = Rectangle(self.mc, config)
        self.assertEqual(widget.anchor_offset_pos, (-10, -10))

    def test_on_container_parent(self):
        parent = self.mc.targets['default'].add_slide(name='parent')

        # test without rounding
        config = {"anchor_x": "center", "anchor_y": "middle",
                  "width": 11, "height": 11, "type": "rectangle"}
        self.mc.config_validator.validate_config('widgets:rectangle', config, base_spec='widgets:common')
        widget = Rectangle(self.mc, config)

        widget.on_container_parent(None, parent)
        self.assertEqual(widget.pos, [400, 300])

        # test with offsetting down
        config = {"anchor_x": "center", "anchor_y": "middle",
                  "width": 11, "height": 11, "type": "rectangle",
                  "round_anchor_x": "left", "round_anchor_y": "bottom",}
        self.mc.config_validator.validate_config('widgets:rectangle', config, base_spec='widgets:common')
        widget = Rectangle(self.mc, config)
        widget.on_container_parent(None, parent)
        self.assertEqual(widget.pos, [399.5, 299.5])

        # test with offsetting up
        config = {"anchor_x": "center", "anchor_y": "middle",
                  "width": 11, "height": 11, "type": "rectangle",
                  "round_anchor_x": "right", "round_anchor_y": "top",}
        self.mc.config_validator.validate_config('widgets:rectangle', config, base_spec='widgets:common')
        widget = Rectangle(self.mc, config)
        widget.on_container_parent(None, parent)
        self.assertEqual(widget.pos, [400.5, 300.5])

        # test with inheriting parent offsets
        config = {"anchor_x": "center", "anchor_y": "middle",
                  "width": 11, "height": 11, "type": "rectangle"}
        parent.display.config['round_anchor_x'] = "left"
        parent.display.config['round_anchor_y'] = "top"
        self.mc.config_validator.validate_config('widgets:rectangle', config, base_spec='widgets:common')
        widget = Rectangle(self.mc, config)
        widget.on_container_parent(None, parent)
        self.assertEqual(widget.pos, [399.5, 300.5])

        # test with widget config overriding parent offset
        config = {"anchor_x": "center", "anchor_y": "middle",
                  "width": 11, "height": 11, "type": "rectangle",
                  "round_anchor_x": "right", "round_anchor_y": "bottom"}
        parent.display.config['round_anchor_x'] = "left"
        parent.display.config['round_anchor_y'] = "top"
        self.mc.config_validator.validate_config('widgets:rectangle', config, base_spec='widgets:common')
        widget = Rectangle(self.mc, config)
        widget.on_container_parent(None, parent)
        self.assertEqual(widget.pos, [400.5, 299.5])

        # test with widget config removing parent offset
        config = {"anchor_x": "center", "anchor_y": "middle",
                  "width": 11, "height": 11, "type": "rectangle",
                  "round_anchor_x": "center", "round_anchor_y": "middle"}
        parent.display.config['round_anchor_x'] = "left"
        parent.display.config['round_anchor_y'] = "top"
        self.mc.config_validator.validate_config('widgets:rectangle', config, base_spec='widgets:common')
        widget = Rectangle(self.mc, config)
        widget.on_container_parent(None, parent)
        self.assertEqual(widget.pos, [400, 300])

    def test_calculate_initial_position(self):
        # Parent is
        # 100x100, so center of the parent is 50, 50

        # test with all defaults

        res_x, res_y = Widget.calculate_initial_position(parent_h=100, parent_w=100,
                                                         x=None, y=None)
        self.assertEqual((res_x, res_y), (50, 50))

        # test positive x, y numbers

        res_x, res_y = Widget.calculate_initial_position(parent_h=100, parent_w=100,
                                                         x=10, y=10)
        self.assertEqual((res_x, res_y), (10, 10))

        res_x, res_y = Widget.calculate_initial_position(parent_h=100, parent_w=100,
                                                         x=33, y=66)
        self.assertEqual((res_x, res_y), (33, 66))

        # test negative x, y numbers

        res_x, res_y = Widget.calculate_initial_position(parent_h=100, parent_w=100,
                                                         x=-10, y=-10)
        self.assertEqual((res_x, res_y), (-10, -10))

        res_x, res_y = Widget.calculate_initial_position(parent_h=100, parent_w=100,
                                                         x=-33, y=-66)
        self.assertEqual((res_x, res_y), (-33, -66))

        # test positive percentages

        res_x, res_y = Widget.calculate_initial_position(parent_h=100, parent_w=100,
                                                         x="80%", y="20%")
        self.assertEqual((res_x, res_y), (80, 20))

        # test negative percentages (dunno how useful these are, but they work)

        res_x, res_y = Widget.calculate_initial_position(parent_h=100, parent_w=100,
                                                         x="-80%", y="-20%")
        self.assertEqual((res_x, res_y), (-80, -20))

        # test positioning strings

        res_x, res_y = Widget.calculate_initial_position(parent_h=100, parent_w=100,
                                                         x="center", y="center")
        self.assertEqual((res_x, res_y), (50, 50))

        res_x, res_y = Widget.calculate_initial_position(parent_h=100, parent_w=100,
                                                         x="center+10", y="center + 10")
        self.assertEqual((res_x, res_y), (60, 60))

        res_x, res_y = Widget.calculate_initial_position(parent_h=100, parent_w=100,
                                                         x="middle-10", y="middle - 10")
        self.assertEqual((res_x, res_y), (40, 40))

        res_x, res_y = Widget.calculate_initial_position(parent_h=100, parent_w=100,
                                                         x="left", y="bottom")
        self.assertEqual((res_x, res_y), (0, 0))

        res_x, res_y = Widget.calculate_initial_position(parent_h=100, parent_w=100,
                                                         x="right", y="top")
        self.assertEqual((res_x, res_y), (100, 100))

    def test_widget_loading_from_config(self):
        # check that all were loaded. First is a dict
        self.assertIn('widget1', self.mc.widgets)
        self.assertIs(type(self.mc.widgets['widget1']), list)

        # List with one item
        self.assertIn('widget2', self.mc.widgets)
        self.assertIs(type(self.mc.widgets['widget2']), list)

        # Lists with multiple items.
        self.assertIn('widget3', self.mc.widgets)
        self.assertIs(type(self.mc.widgets['widget3']), list)
        self.assertEqual(len(self.mc.widgets['widget3']), 3)

        # Ensure they're in order. Order is the order they're drawn. We don't
        # care about z values at this point since those are threaded in when
        # the widgets are added to the slides, but we want to make sure that
        # widgets of the same z are in the order based on their order in the
        # config file.
        self.assertEqual(self.mc.widgets['widget3'][0]['text'],
                         'widget3.1')
        self.assertEqual(self.mc.widgets['widget3'][1]['text'],
                         'widget3.2')
        self.assertEqual(self.mc.widgets['widget3'][2]['text'],
                         'widget3.3')

        # List with multiple items and custom z orders
        self.assertIn('widget4', self.mc.widgets)

    def test_widget_reused_by_name(self):
        self.assertIn('widget_reusable', self.mc.widgets)

        # Named widgets can only be used in modes
        self.mc.modes['mode1'].start()
        self.mc.events.post('show_slide_with_named_widget')
        self.advance_time()
        current_slide = self.mc.targets['default'].current_slide

        self.assertEqual(self.mc.targets['default'].current_slide.name, 'slide_with_named_widget')
        self.assertIn("One Use Widget", [x.widget.text for x in current_slide.widgets])
        self.assertIn("Reusable Widget", [x.widget.text for x in current_slide.widgets])

    def test_widget_with_placeholder(self):
        self.assertIn('widget_placeholder_value1', self.mc.widgets)
        self.assertIn('widget_placeholder_value2', self.mc.widgets)
        self.mc.targets['default'].add_slide(name='blank_slide_one')
        self.mc.targets['default'].add_slide(name='blank_slide_two')
        self.mc.modes['mode1'].start()

        self.mc.targets['default'].show_slide('blank_slide_one')
        self.mc.events.post('show_widget_with_placeholder', value="value1")
        self.advance_time()
        self.assertEqual(["Placeholder widget", "Value One"],
                         [x.widget.text for x in self.mc.targets['default'].current_slide.widgets])

        self.mc.targets['default'].show_slide('blank_slide_two')
        self.mc.events.post('show_widget_with_placeholder', value="value2")
        self.advance_time()
        self.assertEqual(["Placeholder widget", "Value Two"],
                         [x.widget.text for x in self.mc.targets['default'].current_slide.widgets])

    def test_widget_z_order_from_named_widget(self):
        self.mc.targets['default'].add_slide(name='slide1')
        self.mc.targets['default'].show_slide('slide1')
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'slide1')

        self.mc.targets['default'].current_slide.add_widgets_from_library(
                'widget4')

        # initial order & z orders from config
        # 4.1 / 1
        # 4.2 / 1000
        # 4.3 / None
        # 4.4 / 1
        # 4.5 / 1000
        # 4.6 / None
        # 4.7 / None

        # Order should be by z order (highest first), then by order in the
        # config.

        target_order = ['4.2', '4.5', '4.1', '4.4', '4.3', '4.6', '4.7']
        for widget_container, index in zip(
                self.mc.targets['default'].current_slide.widgets,
                target_order):
            self.assertEqual(widget_container.widget.text, 'widget{}'.format(index))

        # add widget 5 which is z order 200
        self.mc.targets['default'].current_slide.add_widgets_from_library(
                'widget5')

        # should be inserted between 4.5 and 4.1
        target_order = ['4.2', '4.5', '5', '4.1', '4.4', '4.3', '4.6', '4.7']
        for widget_container, index in zip(
                self.mc.targets['default'].current_slide.widgets,
                target_order):
            self.assertEqual(widget_container.widget.text, 'widget{}'.format(index))

    def test_widget_z_order_from_slide_player(self):
        self.mc.events.post('show_slide_with_widgets')
        self.advance_time()
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'slide_1')

        # initial order & z orders from config
        # 4.1 / 1
        # 4.2 / 1000
        # 4.3 / None
        # 4.4 / 1
        # 4.5 / 1000
        # 4.6 / None
        # 4.7 / None

        # Order should be by z order (highest first), then by order in the
        # config.

        target_order = ['4.2', '4.5', '4.1', '4.4', '4.3', '4.6', '4.7']
        for widget_container, index in zip(
                self.mc.targets['default'].current_slide.widgets,
                target_order):
            self.assertEqual(widget_container.widget.text, 'widget{}'.format(index))

    def test_widget_z_order_from_named_slide(self):
        self.mc.events.post('show_slide_with_lots_of_widgets')
        self.advance_time()
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'slide_with_lots_of_widgets')

        # initial order & z orders from config
        # 4.1 / 1
        # 4.2 / 1000
        # 4.3 / None
        # 4.4 / 1
        # 4.5 / 1000
        # 4.6 / None
        # 4.7 / None

        # Order should be by z order (highest first), then by order in the
        # config.

        target_order = ['4.2', '4.5', '4.1', '4.4', '4.3', '4.6', '4.7']
        for widget_container, index in zip(
                self.mc.targets['default'].current_slide.widgets,
                target_order):
            self.assertEqual(widget_container.widget.text, 'widget{}'.format(index))

    def test_widget_player_add_to_current_slide(self):
        # create a slide
        self.mc.targets['default'].add_slide(name='slide1')
        self.mc.targets['default'].show_slide('slide1')
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'slide1')

        # post the event to add widget1 to the default target, default slide
        self.mc.events.post('add_widget1_to_current')
        self.advance_time()

        # widget1 should be on the default slide
        self.assertIn('_global-widget1', [x.widget.key for x in self.mc.targets[
            'default'].current_slide.widgets])

    def test_widget_player_add_to_named_slide(self):
        # create two slides
        self.mc.targets['default'].add_slide(name='slide1')
        self.mc.targets['default'].show_slide('slide1')
        self.advance_time()
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'slide1')

        self.mc.targets['default'].add_slide(name='slide2')
        self.mc.targets['default'].show_slide('slide2')
        self.advance_time()
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'slide2')

        # Add widget2 to slide 1
        self.mc.events.post('add_widget2_to_slide1')
        self.advance_time()

        # widget2 should be in slide1, not slide2, not current slide
        self.assertIn('widget2',
                      [x.widget.text for x in
                       self.mc.active_slides['slide1'].widgets])
        self.assertNotIn('widget2',
                         [x.widget.text for x in
                          self.mc.active_slides['slide2'].widgets])
        self.assertNotIn('widget2', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])

        # Remove widget2
        self.mc.events.post('remove_widget2')
        self.advance_time()

        # widget2 should not be present in slide1
        self.assertNotIn('widget2',
                         [x.widget.text for x in
                          self.mc.active_slides['slide1'].widgets])

        # Update widget2 on slide1
        self.mc.events.post('update_widget2')
        self.advance_time()

        # widget2 should be in slide1, not slide2, not current slide
        self.assertIn('widget2',
                      [x.widget.text for x in
                       self.mc.active_slides['slide1'].widgets])
        self.assertNotIn('widget2',
                         [x.widget.text for x in
                          self.mc.active_slides['slide2'].widgets])
        self.assertNotIn('widget2', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])

        # show slide1 and make sure the widget is there
        self.mc.targets['default'].current_slide_name = 'slide1'
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'slide1')
        self.assertIn('widget2', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])

    def test_widget_player_add_to_invalid_slide(self):
        self.mc.targets['default'].add_slide(name='slide2')
        self.mc.targets['default'].show_slide('slide2')
        self.advance_time()
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'slide2')

        self.mc.events.post('add_widget2_to_slide1')

        with self.assertRaises(Exception) as e:
            self.advance_time()
        self.assertIsInstance(e.exception.__cause__, KeyError)

    def test_widget_player_with_different_key_than_named_widget(self):
        self.mc.targets['default'].add_slide(name='slide1')
        self.mc.targets['default'].show_slide('slide1')
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'slide1')

        # widget_player for show_widget9 specifies widget9 with a key
        # "widget9_wp_key", but widget9 in the widgets: section has a key
        # "widget9_key"

        self.mc.events.post('show_widget9')

        with self.assertRaises(Exception) as e:
            self.advance_time()
        self.assertIsInstance(e.exception.__cause__, KeyError)
        self.assertEqual(str(e.exception.__cause__), "\"Widget has incoming key 'wigdet9_wp_key' which does not "
                                                     "match the key in the widget's config 'widget9_key'.\"")

    def test_removing_mode_widget_on_mode_stop(self):
        # create a slide and add some base widgets
        self.mc.targets['default'].add_slide(name='slide1')
        self.mc.targets['default'].show_slide('slide1')
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'slide1')

        self.mc.events.post('add_widget1_to_current')
        self.advance_time()

        # verify widget 1 is there but not widget 2
        self.assertIn('widget1', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])
        self.assertNotIn('widget2', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])

        # start a mode
        self.mc.modes['mode1'].start()
        self.advance_time()

        # post the event to add the widget. This will also test that the
        # widget_player in a mode can add a widget from the base
        self.mc.events.post('mode1_add_widgets')
        self.advance_time()

        # make sure the new widget is there, and the old one is still there
        self.assertIn('widget1', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])
        self.assertIn('widget2', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])

        # stop the mode
        self.mc.modes['mode1'].stop()
        self.advance_time()

        # make sure the mode widget is gone, but the first one is still there
        self.assertIn('widget1', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])
        self.assertNotIn('widget2', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])

    def test_removing_mode_widget_with_custom_key_on_mode_stop(self):
        # create a slide and add some base widgets
        self.mc.targets['default'].add_slide(name='slide1')
        self.mc.targets['default'].show_slide('slide1')
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'slide1')

        self.mc.events.post('add_widget1_to_current')
        self.advance_time()

        # verify widget 1 is there but not widget 2
        self.assertIn('widget1', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])
        self.assertNotIn('widget2', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])

        # start a mode
        self.mc.modes['mode1'].start()
        self.advance_time()

        # post the event to add the widget. This will also test that the
        # widget_player in a mode can add a widget from the base
        self.mc.events.post('mode1_add_widget_with_key')
        self.advance_time()

        # make sure the new widget is there, and the old one is still there
        self.assertIn('widget1', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])
        self.assertIn('widget2', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])

        # make sure the key of the new widget is correct
        widget2 = self.mc.targets[
            'default'].current_slide.widgets[1].widget
        self.assertEqual(widget2.text, 'widget2')
        self.assertEqual(widget2.key, 'newton_crosby')

        # stop the mode
        self.mc.modes['mode1'].stop()
        self.advance_time()

        # make sure the mode widget is gone, but the first one is still there
        self.assertIn('widget1', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])
        self.assertNotIn('widget2', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])

    def test_widgets_in_slide_frame_parent(self):
        # create a slide and add some base widgets
        self.mc.targets['default'].add_slide(name='slide1')
        self.mc.targets['default'].show_slide('slide1')
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'slide1')

        self.mc.events.post('add_widget1_to_current')
        self.advance_time()

        # verify widget 1 is there but not widget 6
        self.assertIn('widget1', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])
        self.assertNotIn('widget6', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])

        # add widget 6, z: -1, so it should go in slide parent
        self.mc.events.post('add_widget6')
        self.advance_time()

        # verify widget1 is in the slide but not widget 6
        self.assertIn('widget1', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])
        self.assertNotIn('widget6', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])

        # verify widget6 is the highest priority in the parent frame
        # (highest priority is the first element in the list)
        self.assertEqual('widget6', self.mc.targets[
            'default'].parent_widgets[0].widget.text)

        # now switch the slide
        self.mc.targets['default'].add_slide(name='slide2')
        self.mc.targets['default'].show_slide('slide2')
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'slide2')

        # make sure neither widget is in this slide
        self.assertNotIn('widget1', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])
        self.assertNotIn('widget6', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])

        # make sure widget6 is still in the SlideFrameParent
        self.assertEqual('widget6', self.mc.targets[
            'default'].parent_widgets[0].widget.text)

    def test_widget_to_parent_via_widget_settings(self):
        # create a slide and add some base widgets
        self.mc.targets['default'].add_slide(name='slide1')
        self.mc.targets['default'].show_slide('slide1')
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'slide1')

        self.mc.events.post('add_widget1_to_current')
        self.advance_time()

        # verify widget 1 is there but not box11
        self.assertIn('widget1', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])
        self.assertNotIn('box11', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])
        self.assertNotIn('box12', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])

        # add box11 into parent and box12 into slide
        self.mc.events.post('widget_to_parent')
        self.advance_time()

        # verify widget1 and box12 are in the slide but not box11
        self.assertIn('widget1', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])
        self.assertNotIn('box11', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])
        self.assertIn('box12', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])

        # verify box11 is in the parent frame
        self.assertTrue(isinstance(self.mc.targets[
            'default'].parent_widgets[0], WidgetContainer))

        self.assertEqual('box11', self.mc.targets[
            'default'].parent_widgets[0].widget.text)

        # switch the slide
        self.mc.events.post('show_new_slide')
        self.advance_time()
        self.assertEqual('box11', self.mc.targets[
            'default'].parent_widgets[0].widget.text)
        self.assertEqual('NEW SLIDE', self.mc.targets[
            'default'].current_slide.widgets[0].widget.text)

        # make sure positioning works
        self.assertEqual(0, self.mc.targets[
            'default'].current_slide.widgets[0].widget.y)

    def test_removing_mode_widget_from_parent_frame_on_mode_stop(self):
        # create a slide and add some base widgets
        self.mc.targets['default'].add_slide(name='slide1')
        self.mc.targets['default'].show_slide('slide1')
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'slide1')

        self.mc.events.post('add_widget1_to_current')
        self.advance_time()

        # verify widget 1 is there but not widget 6
        self.assertIn('widget1', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])
        self.assertNotIn('widget6', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])

        # start a mode
        self.mc.modes['mode1'].start()
        self.advance_time()

        # post the event to add the widget to the slide parent
        self.mc.events.post('mode1_add_widget6')
        self.advance_time()

        # make sure the new widget is not in the slide, but the old one is
        self.assertIn('widget1', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])
        self.assertNotIn('widget6', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])

        # verify widget6 is the highest priority in the parent frame
        self.assertEqual('widget6', self.mc.targets[
            'default'].parent_widgets[0].widget.text)
        self.assertTrue(isinstance(self.mc.targets[
            'default'].parent_widgets[0].widget, Text))

        # stop the mode
        self.mc.modes['mode1'].stop()
        self.advance_time()

        # make sure the the first one is still there
        self.assertIn('widget1', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])

        # verify widget6 is gone
        self.assertNotIn('widget6', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.parent_widgets])

    def test_removing_widget(self):
        self.mc.targets['default'].add_slide(name='slide1')
        self.mc.targets['default'].show_slide('slide1')
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'slide1')

        # post the event to add widget1 to the default target, default slide
        self.mc.events.post('add_widget1_to_current')
        self.mc.events.post('add_widget2_to_current')
        self.advance_time()

        self.assertIn('widget1', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])
        self.assertIn('widget2', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])

        self.mc.events.post('remove_widget1')
        self.advance_time()

        self.assertNotIn('widget1', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])
        self.assertIn('widget2', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])

    def test_widget_expire(self):
        self.mc.targets['default'].add_slide(name='slide1')
        self.mc.targets['default'].show_slide('slide1')
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'slide1')

        self.mc.events.post('add_widget1_to_current')
        self.mc.events.post('add_widget7')
        self.advance_time()
        widget7 = weakref.ref([x.widget for x in self.mc.targets['default'].current_slide.widgets
                               if x.widget.key == '_global-widget7'][0])
        self.assertTrue(widget7())

        self.assertIn('_global-widget1', [x.widget.key for x in self.mc.targets[
            'default'].current_slide.widgets])
        self.assertIn('_global-widget7', [x.widget.key for x in self.mc.targets[
            'default'].current_slide.widgets])

        self.advance_time(1)

        self.assertIn('_global-widget1', [x.widget.key for x in self.mc.targets[
            'default'].current_slide.widgets])
        self.assertNotIn('_global-widget7', [x.widget.key for x in self.mc.targets[
            'default'].current_slide.widgets])

        gc.collect()
        self.assertFalse(widget7())

    def test_widget_player_expire(self):
        self.mc.targets['default'].add_slide(name='slide1')
        self.mc.targets['default'].show_slide('slide1')
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'slide1')

        self.mc.events.post('add_widget1_to_current')
        self.mc.events.post('add_widget8_expire')
        self.advance_time()

        self.assertIn('_global-widget1', [x.widget.key for x in self.mc.targets[
            'default'].current_slide.widgets])
        self.assertIn('_global-widget8', [x.widget.key for x in self.mc.targets[
            'default'].current_slide.widgets])

        widget8 = weakref.ref([x.widget for x in self.mc.targets['default'].current_slide.widgets
                               if x.widget.key == '_global-widget8'][0])
        self.assertTrue(widget8())

        self.advance_time(1)

        self.assertIn('_global-widget1', [x.widget.key for x in self.mc.targets[
            'default'].current_slide.widgets])
        self.assertNotIn('_global-widget8', [x.widget.key for x in self.mc.targets[
            'default'].current_slide.widgets])

        gc.collect()
        self.assertFalse(widget8())

    def test_widget_player_expire_in_parent_frame(self):
        self.mc.targets['default'].add_slide(name='slide1')
        self.mc.targets['default'].show_slide('slide1')
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'slide1')

        self.mc.events.post('add_widget1_to_current')
        self.mc.events.post('add_widget8_expire_parent')
        self.advance_time()

        self.assertIn('_global-widget1', [x.widget.key for x in self.mc.targets[
            'default'].current_slide.widgets])
        self.assertIn('_global-widget8', [x.widget.key for x in self.mc.targets[
            'default'].parent_widgets])

        widget8 = weakref.ref(self.mc.targets[
            'default'].parent_widgets[0])
        self.assertTrue(widget8())

        self.advance_time(1)

        self.assertIn('_global-widget1', [x.widget.key for x in self.mc.targets[
            'default'].current_slide.widgets])
        self.assertNotIn('_global-widget8', [x.key for x in self.mc.targets[
            'default'].parent.walk()])

        gc.collect()
        self.assertFalse(widget8())

    def test_widget_player_custom_widget_settings(self):
        self.mc.targets['default'].add_slide(name='slide1')
        self.mc.targets['default'].show_slide('slide1')
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'slide1')

        self.mc.events.post('add_widget8_custom_settings')
        self.advance_time()

        w8 = [x.widget for x in self.mc.targets[
              'default'].current_slide.widgets
              if x.widget.key == '_global-widget8'][0]

        self.assertEqual([1, 0, 0, 1], w8.color)
        self.assertEqual(70, w8.font_size)
        self.assertAlmostEqual(-310, w8.anchor_offset_pos[0], delta=20)
        self.assertEqual(790, w8.x)  # anchor_x: right, x: right-10

    def test_widget_removal_from_slide_player(self):
        # tests that we can remove a widget by key that was shown via the
        # slide player instead of the widget player
        self.mc.events.post('show_slide_1')
        self.advance_time()

        # make sure the two widgets are there
        self.assertIn('WIDGET WITH KEY', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])
        self.assertIn('WIDGET NO KEY', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])

        widget = weakref.ref([x.widget for x in self.mc.targets['default'].current_slide.widgets
                               if x.widget.text == 'WIDGET WITH KEY'][0])
        self.assertTrue(widget())

        self.mc.events.post('remove_widget1_by_key')
        self.advance_time()

        # make sure the one with key is gone but the other is there
        self.assertNotIn('WIDGET WITH KEY', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])
        self.assertIn('WIDGET NO KEY', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])

        gc.collect()
        self.assertFalse(widget())

    def test_widget_expire_from_slide_player(self):
        # tests that we can remove a widget by key that was shown via the
        # slide player instead of the widget player
        self.mc.events.post('show_slide_1_with_expire')
        self.advance_time()

        # make sure the two widgets are there
        self.assertIn('WIDGET EXPIRE 1s', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])
        self.assertIn('WIDGET NO EXPIRE', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])

        widget = weakref.ref([x.widget for x in self.mc.targets['default'].current_slide.widgets
                               if x.widget.text == 'WIDGET EXPIRE 1s'][0])
        self.assertTrue(widget())

        self.advance_time(1)

        # make sure the one with key is gone but the other is there
        self.assertNotIn('WIDGET EXPIRE 1s', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])
        self.assertIn('WIDGET NO EXPIRE', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])

        gc.collect()
        self.assertFalse(widget())

    def test_opacity(self):
        self.mc.targets['default'].add_slide(name='slide1')
        self.mc.targets['default'].show_slide('slide1')
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'slide1')

        self.mc.events.post('add_widget8_opacity_50')
        self.advance_time()

        w8 = [x.widget for x in self.mc.targets['default'].current_slide.widgets
              if x.widget.key == '_global-widget8'][0]

        self.assertEqual(.5, w8.opacity)

    def test_updating_widget_settings(self):
        self.mc.events.post('show_slide_2')
        self.advance_time()

        self.mc.events.post('event_a')
        self.advance_time()

        widget = self.mc.targets['default'].current_slide.find_widgets_by_key(
            '_global-widget1')[0]
        self.assertEqual(widget.text, 'A')
        self.assertEqual(widget.color, [1.0, 0.0, 0.0, 1.0])

        self.mc.events.post('event_s')
        self.advance_time()

        old_widget = weakref.ref(widget)

        widget = self.mc.targets['default'].current_slide.find_widgets_by_key(
            '_global-widget1')[0]
        self.assertEqual(widget.text, 'S')
        self.assertEqual(widget.color, [0.0, 1.0, 0.0, 1.0])

        self.mc.events.post('event_d')
        self.advance_time()

        gc.collect()
        self.assertFalse(old_widget())

        old_widget = weakref.ref(widget)

        widget = self.mc.targets['default'].current_slide.find_widgets_by_key(
            '_global-widget1')[0]
        self.assertEqual(widget.text, 'D')
        self.assertEqual(widget.color, [0.0, 0.0, 1.0, 1.0])

        gc.collect()
        self.assertFalse(old_widget())

    def test_widget_updating(self):
        self.mc.events.post('show_slide_3')
        self.advance_time()

        self.mc.events.post('widget_4up')
        self.advance_time()

        self.mc.events.post('widget_4up_red')
        self.advance_time()

    def test_widget_with_key(self):
        self.mc.targets['default'].add_slide(name='slide1')
        self.mc.targets['default'].show_slide('slide1')
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'slide1')

        self.assertNotIn('widget2', [x.text for x in self.mc.targets[
            'default'].current_slide.widgets])

        # post the event to add widget2 to slide1
        self.mc.events.post('show_christmas_slide_full')
        self.advance_time()

        # should be there
        self.assertIn('widget2', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])

        # remove widget2 again
        self.mc.events.post('remove_christmas_full')
        self.advance_time()

        # should no longer be there
        self.assertNotIn('widget2', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])

        # remove slide
        self.mc.targets['default'].remove_slide('slide1')

        # post the event to add widget2 to slide1
        self.mc.events.post('show_christmas_slide_full')
        self.advance_time()

        # slide1 is not there. widget2 should also not be there
        self.assertNotIn('widget2', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])

        # show slide
        self.mc.targets['default'].show_slide('slide1')
        self.advance_time()

        # should be there (automagically)
        self.assertIn('widget2', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])

        # remove slide
        self.mc.targets['default'].remove_slide('slide1')

        # slide1 is not there. widget2 should also not be there
        self.assertNotIn('widget2', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])

        # show slide
        self.mc.targets['default'].show_slide('slide1')
        self.advance_time()

        # should be there (still)
        self.assertIn('widget2', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])

        # remove widget2 again
        self.mc.events.post('remove_christmas_full')
        self.advance_time()

        # should no longer be there
        self.assertNotIn('widget2', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])

        # remove slide
        self.mc.targets['default'].remove_slide('slide1')

        # post the event to add widget2 to slide1
        self.mc.events.post('show_christmas_slide_full')
        self.advance_time()

        # slide1 is not there. widget2 should also not be there
        self.assertNotIn('widget2', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])

        # remove widget2 again
        self.mc.events.post('remove_christmas_full')
        self.advance_time()

        # show slide
        self.mc.targets['default'].show_slide('slide1')
        self.advance_time()

        # should not appear
        self.assertNotIn('widget2', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])

    def test_updating_mode_widget_by_key(self):
        # create a slide and add some base widgets
        self.mc.targets['default'].add_slide(name='slide1')
        self.mc.targets['default'].show_slide('slide1')
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'slide1')
        self.mc.events.post('add_widget1_to_current')
        self.advance_time()

        # verify widget 1 is there but not widget 2
        self.assertIn('widget1', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])
        self.assertNotIn('widget2', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])

        # start a mode
        self.mc.modes['mode1'].start()
        self.advance_time()

        # post the event to add the widget. This will also test that the
        # widget_player in a mode can add a widget from the base
        self.mc.events.post('mode1_add_widget_with_key')
        self.advance_time()

        # make sure the new widget is there, and the old one is still there
        self.assertIn('widget1', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])
        self.assertIn('widget2', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])

        # make sure the key of the new widget is correct
        widget2 = self.mc.targets[
            'default'].current_slide.widgets[1].widget
        self.assertEqual(widget2.text, 'widget2')
        self.assertEqual(widget2.key, 'newton_crosby')

        # update widget2 by key
        self.mc.events.post('mode1_update_widget2')
        self.advance_time()

        widget2 = self.mc.targets[
            'default'].current_slide.widgets[1].widget
        self.assertEqual(widget2.text, 'UPDATED TEXT')
        self.assertEqual(widget2.key, 'newton_crosby')

    def test_widget_player_with_placeholder(self):
        self.mc.targets['default'].add_slide(name='slide1')
        self.mc.targets['default'].show_slide('slide1')
        self.mc.events.post('show_widget10', text="asd")
        self.advance_time()

        # verify asd is there
        self.assertIn('asd', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])

    def test_slide_frame_widget(self):
        self.mc.events.post("show_info_frame")
        self.advance_time(1)

    def test_properties_from_named_widget(self):
        self.mc.events.post('add_widget1_to_current')
        self.mc.events.post('add_widget2_to_current')
        self.advance_time()

        widget1 = self.mc.targets[
            'default'].current_slide.widgets[0].widget

        widget2 = self.mc.targets[
            'default'].current_slide.widgets[1].widget

        self.assertEqual(widget1.text, "widget1")
        self.assertEqual(widget2.text, "widget2")

        self.assertEqual(widget1.y, 360)
        self.assertEqual(widget2.y, 50)
        self.assertEqual(widget1.font_size, 100)
        self.assertEqual(widget2.font_size, 100)
        self.assertEqual(widget1.color, [1.0, 1.0, 0.0, 1])
        self.assertEqual(widget2.color, [1.0, 0.0, 0.0, 1])

    def test_animation_properties_from_widget_player(self):
        """Test multiple widget types with animations."""

        self.mc.events.post('show_text_widget')
        self.mc.events.post('show_bezier_widget')
        self.mc.events.post('show_rectangle_widget')
        self.mc.events.post('show_line_widget')
        self.mc.events.post('show_ellipse_widget')
        self.advance_real_time(1.5)

        self.mc.events.post('show_ellipse_widget')
        self.mc.events.post('show_quad_widget')
        self.mc.events.post('show_points_widget')
        self.mc.events.post('show_triangle_widget')
        self.advance_real_time(4.3)

        text_widget = self.mc.targets['default'].current_slide.widgets[0].widget
        self.assertIsInstance(text_widget, Text)
        self.assertEqual(text_widget.rotation, 45)
        self.assertEqual(text_widget.scale, 0.75)

        bezier_widget = self.mc.targets['default'].current_slide.widgets[1].widget
        self.assertIsInstance(bezier_widget, Bezier)
        self.assertEqual(bezier_widget.color, [1, 1, 0, 1])
        self.assertEqual(bezier_widget.rotation, -300)
        self.assertEqual(bezier_widget.points, [200, 200, 50, 100, 100, 250])

        rectangle_widget = self.mc.targets['default'].current_slide.widgets[2].widget
        self.assertIsInstance(rectangle_widget, Rectangle)
        self.assertEqual(rectangle_widget.rotation, 0)
        self.assertEqual(rectangle_widget.scale, 1)

        line_widget = self.mc.targets['default'].current_slide.widgets[3].widget
        self.assertIsInstance(line_widget, Line)
        self.assertEqual(line_widget.rotation, 360)
        self.assertEqual(line_widget.scale, 1.5)

        ellipse_widget = self.mc.targets['default'].current_slide.widgets[4].widget
        self.assertIsInstance(ellipse_widget, Ellipse)
        self.assertEqual(ellipse_widget.rotation, 360)
        self.assertEqual(ellipse_widget.pos, [500, 400])

        quad_widget = self.mc.targets['default'].current_slide.widgets[5].widget
        self.assertIsInstance(quad_widget, Quad)
        self.assertEqual(quad_widget.rotation, 0)
        self.assertEqual(quad_widget.scale, 1)
        self.assertEqual(quad_widget.points, [300, 100, 350, 200, 500, 150, 450, 50])

        points_widget = self.mc.targets['default'].current_slide.widgets[6].widget
        self.assertIsInstance(points_widget, Point)
        self.assertEqual(points_widget.rotation, 900)
        self.assertEqual(points_widget.scale, 1.5)
        self.assertEqual(points_widget.points, [100, 450, 100, 550, 200, 450])
        self.assertEqual(points_widget.pointsize, 8)

        triangle_widget = self.mc.targets['default'].current_slide.widgets[7].widget
        self.assertIsInstance(triangle_widget, Triangle)
        self.assertEqual(triangle_widget.rotation, -900)
        self.assertEqual(triangle_widget.scale, 1.5)
        self.assertEqual(triangle_widget.points, [100, 450, 100, 550, 200, 450])

    def test_events_when_removed(self):
        """Test events_when_removed property to ensure custom events are posted."""

        # Mock BCP send method
        self.mc.bcp_processor.send = MagicMock()
        self.mc.bcp_processor.enabled = True

        self.mc.events.post('show_custom_events1_widget')
        self.advance_real_time()
        self.mc.bcp_processor.send.assert_any_call('trigger', name='custom_events1_added')

        self.mc.events.post('show_custom_events2_widget')
        self.advance_real_time()
        self.mc.bcp_processor.send.assert_any_call('trigger', name='custom_events2_added')
        self.mc.bcp_processor.send.assert_any_call('trigger', name='custom_events2_added_again')

        self.mc.events.post('remove_custom_events1_widget')
        self.advance_real_time(0.1)
        self.mc.bcp_processor.send.assert_any_call('trigger', name='custom_events1_removed')

        self.mc.events.post('remove_custom_events2_widget')
        self.advance_real_time()
        self.mc.bcp_processor.send.assert_any_call('trigger', name='custom_events2_removed')
        self.mc.bcp_processor.send.assert_any_call('trigger', name='custom_events2_removed_again')

        self.mc.bcp_processor.send.reset_mock()
        self.mc.events.post('show_new_slide')
        self.advance_real_time()
        self.mc.bcp_processor.send.assert_any_call('trigger', name='text_on_new_slide2_added')

        self.mc.events.post('show_slide_1')
        self.mc.events.post('remove_new_slide')
        self.advance_real_time()
        self.mc.bcp_processor.send.assert_any_call('trigger', name='text_on_new_slide2_removed')





from copy import deepcopy
from mpfmc.core.mc_config_player import McConfigPlayer


class McSlidePlayer(McConfigPlayer):
    """Base class for the Slide Player that runs on the mpf-mc side of things.
    It receives all of its instructions via BCP from a MpfSlidePlayer instance
    running as part of MPF.

    The slide_player: section of a config file (either the machine-wide or
    a mode-specific config) looks like this:

    slide_player:
        <event_name>:
            <slide_name>:
                <slide_settings>: ...

    The express config just puts a slide_name next to an event. (In this case
    it assumes the slide has already been defined in yours slides: section)

    slide_player:
        some_event: slide_name_to_show

    If you want to control other settings (such as a display target, priority,
    transition, etc.), enter the slide name on the next line and the settings
    indented under it, like this:

    slide_player:
        some_event:
            slide_name_to_show:
                target: dmd
                transition: fade

    Alternately you can pass a full configuration. You can also pass a widget
    or list of widgets to create the slide if based on the settings in the
    slide_player. Here are several various examples:

    slide_player:
        some_event:
            slide1:
                transition: move_in

        some_event2:
            slide2:
                transition:
                    type: move_in
                    direction: right
                    duration: 1s

        some_event3: slide3

        some_event4:
            slide4:
                type: text
                text: SOME TEXT
                color: red
                y: 50

        some_event5:
            slide5:
              - type: text
                text: SOME TEXT
                color: red
                y: 50
              - type: text
                text: AND MORE TEXT
                color: blue
                y: 150

        some_event6:
            slide6:
                widgets:
                - type: text
                  text: SOME TEXT
                  color: red
                  y: 50
                - type: text
                  text: AND MORE TEXT
                  color: blue
                  y: 150

        some_event7:
            slide7:
                widgets:
                - type: text
                  text: SOME TEXT
                  color: red
                  y: 50
                - type: text
                  text: AND MORE TEXT
                  color: blue
                  y: 150
                transition: move_in

    """
    config_file_section = 'slide_player'
    show_section = 'slides'
    machine_collection_name = None  # slides are not real devices currently

    def play(self, settings, context, calling_context, priority=0, **kwargs):
        """Process a slide_player event."""
        instance_dict = self._get_instance_dict(context)
        full_context = self._get_full_context(context)

        self.machine.log.info("SlidePlayer: Play called with settings=%s", settings)

        settings = settings['slides'] if 'slides' in settings else settings

        for slide, s in settings.items():
            if not isinstance(slide, str):
                if slide.condition and not slide.condition.evaluate(kwargs):
                    continue
                slide = slide.name

            s = deepcopy(s)
            s.update(kwargs)

            if s.get("slide"):
                slide = s['slide']
            elif slide == "widgets":
                # name of anonymous slides depends on context + event name
                slide = "{}-{}".format(full_context, calling_context)

            try:
                s['priority'] += priority
            except (KeyError, TypeError):
                s['priority'] = priority

            if s['target']:
                target_name = s['target']
            else:
                target_name = 'default'

            if target_name not in self.machine.targets or not self.machine.targets[target_name].ready:
                # Target does not exist yet or is not ready. Perform action when it is ready
                raise AssertionError("Target {} not ready".format(target_name))

            target = self.machine.targets[target_name]

            # remove target
            s.pop("target")

            if target_name not in instance_dict:
                instance_dict[target_name] = {}

            if s['action'] == 'play':
                # remove slide if it already exists
                if slide in instance_dict[target_name]:
                    instance_dict[target_name][slide].remove()
                    del instance_dict[target_name][slide]

                # is this a named slide, or a new slide?
                self.machine.log.debug("SlidePlayer: Playing slide '%s' on target '%s' (Args=%s)",
                                       slide,
                                       target.name,
                                       s)
                if 'widgets' in s:
                    target.add_and_show_slide(key=full_context,
                                              slide_name=slide, play_kwargs=s['tokens'], **s)
                else:
                    target.show_slide(slide_name=slide, key=full_context, play_kwargs=s['tokens'], **s)

                target.get_screen(slide).on_slide_play()

                instance_dict[target_name][slide] = target.get_slide(slide)

            elif s['action'] == 'remove' and slide in instance_dict[target_name]:
                del instance_dict[target_name][slide]
                target.remove_slide(slide=slide,
                                    transition_config=s['transition'] if 'transition' in s else [])

    def _expand_device(self, device):
        # parse conditionals
        devices = super()._expand_device(device)
        for index, device_entry in enumerate(devices):
            if isinstance(device_entry, str):
                devices[index] = self.machine.placeholder_manager.parse_conditional_template(device_entry)
        return devices

    def get_express_config(self, value):
        # express config for slides can either be a string (slide name) or a
        # list (widgets which are put on a new slide)
        if isinstance(value, list):
            return dict(widgets=value)
        else:
            return dict(slide=value)

    def validate_config(self, config):
        """Validates the slide_player: section of a config file (either a
        machine-wide config or a mode config).

        Args:
            config: A dict of the contents of the slide_player section
            from the config file. It's assumed that keys are event names, and
            values are settings for what the slide_player should do when that
            event is posted.

        Returns: A dict a validated entries.

        This method overrides the base method since the slide_player has
        unique options (including lists of widgets or single dict entries that
        are a widget settings instead of slide settings.

        """
        # first, we're looking to see if we have a string, a list, or a dict.
        # if it's a dict, we look to see whether we have a widgets: entry
        # or the name of some slide

        validated_config = dict()

        for event, settings in config.items():
            validated_config[event] = dict()
            validated_config[event]['slides'] = dict()

            if isinstance(settings, list):
                raise AssertionError(
                    "Config of slide_player for event {} is broken. "
                    "It expects a dict not a list".format(event))

            if not isinstance(settings, dict):
                settings = {settings: dict()}

            for slide, slide_settings in settings.items():
                dict_is_widgets = False

                # if settings is list, it's widgets
                if isinstance(slide_settings, list):
                    # convert it to a dict by moving this list of settings into
                    # a dict key called "widgets"
                    slide_settings = dict(widgets=slide_settings)

                # Now we have a dict, but is this a dict of settings for a
                # single slide, or a dict of settings for the slide player
                # itself?

                # So let's check the keys and see if they're all valid keys
                # for a slide_player. If so, it's slide_player keys. If not,
                # we assume they're widgets for a slide.

                elif isinstance(slide_settings, str):
                    # slide_settings could be a string 'slide: slide_name',
                    # so we rename the key to the slide name with an empty dict
                    if slide_settings == "remove":
                        slide_settings = {"action": "remove"}
                    else:
                        slide = slide_settings
                        slide_settings = dict()

                elif not isinstance(slide_settings, dict):
                    raise AssertionError(
                        "Expected a dict in slide_player {}:{}.".format(event,
                                                                        slide))

                for key in slide_settings:
                    if key not in self.machine.config_validator.get_config_spec()['slide_player']:
                        dict_is_widgets = True
                        break

                if dict_is_widgets:
                    slide_settings = dict(widgets=[slide_settings])

                validated_config[event]['slides'].update(
                    self._validate_config_item(slide, slide_settings))

        return validated_config

    def _validate_config_item(self, device, device_settings):
        validated_dict = super()._validate_config_item(device, device_settings)
        # device is slide name
        for config in validated_dict.values():
            if config['target'] and config['target'] not in self.machine.targets:
                raise AssertionError("Target {} does not exist.".format(config['target']))
            if 'widgets' in config:
                # verify widgets
                config['widgets'] = self.machine.widgets.process_config(
                    config['widgets'])
            else:
                # verify that slide exists
                if config['slide']:
                    slide = config['slide']
                else:
                    slide = self.machine.placeholder_manager.parse_conditional_template(device).name

                if config['action'] == "play" and slide not in self.machine.slides:
                    raise AssertionError("Slide {} not found".format(slide))

            self.machine.transition_manager.validate_transitions(config)

        return validated_dict

    def clear_context(self, context):
        """Remove all slides from this player context."""
        instance_dict = self._get_instance_dict(context)
        for _, slides in instance_dict.items():
            for slide in slides.values():
                slide.remove()

        self._reset_instance_dict(context)


mc_player_cls = McSlidePlayer



from mpfmc.tests.MpfMcTestCase import MpfMcTestCase


class TestBitmapFonts(MpfMcTestCase):
    def get_machine_path(self):
        return 'tests/machine_files/bitmap_fonts'

    def get_config_file(self):
        return 'test_bitmap_fonts.yaml'

    def test_loading_asset(self):
        # test that the bitmap_fonts asset class gets built correctly
        self.assertTrue(hasattr(self.mc, 'bitmap_fonts'))

        # Monospaced font with simple descriptor list
        self.assertIn('F1fuv', self.mc.bitmap_fonts)
        f1fuv_font = self.mc.bitmap_fonts['F1fuv']
        self.assertIsNotNone(f1fuv_font)
        self.assertIsNotNone(f1fuv_font.bitmap_font)

        # Test the font descriptor list calculations
        self.assertEqual(f1fuv_font.bitmap_font.scale_w, 801)
        self.assertEqual(f1fuv_font.bitmap_font.scale_h, 300)
        self.assertEqual(f1fuv_font.bitmap_font.line_height, 50)
        self.assertEqual(f1fuv_font.bitmap_font.base, 50)

        self.assertTrue(len(f1fuv_font.bitmap_font.get_characters()), 95)
        self.assertEqual(f1fuv_font.bitmap_font.get_characters()[57].id, 57)
        self.assertEqual(f1fuv_font.bitmap_font.get_characters()[57].rect["w"], 50)
        self.assertEqual(f1fuv_font.bitmap_font.get_characters()[57].rect["h"], 50)
        self.assertEqual(f1fuv_font.bitmap_font.get_characters()[57].xadvance, 50)
        self.assertEqual(f1fuv_font.bitmap_font.get_characters()[57].xoffset, 0)
        self.assertEqual(f1fuv_font.bitmap_font.get_characters()[57].yoffset, 0)
        self.assertEqual(f1fuv_font.bitmap_font.get_characters()[57].rect["x"], 450)
        self.assertEqual(f1fuv_font.bitmap_font.get_characters()[57].rect["y"], 50)

        self.assertEqual(f1fuv_font.bitmap_font.get_characters()[90].rect["x"], 500)
        self.assertEqual(f1fuv_font.bitmap_font.get_characters()[90].rect["y"], 150)

        self.assertEqual(f1fuv_font.bitmap_font.get_characters()[122].rect["x"], 500)
        self.assertEqual(f1fuv_font.bitmap_font.get_characters()[122].rect["y"], 250)

        # Test the extent calculations
        self.assertEqual(f1fuv_font.bitmap_font.get_extents("testing"), (350, 50))
        self.assertEqual(f1fuv_font.bitmap_font.get_extents("more testing"), (600, 50))
        self.assertEqual(f1fuv_font.bitmap_font.get_ascent(), 50)
        self.assertEqual(f1fuv_font.bitmap_font.get_descent(), 0)

        # Variable width font (with xml descriptor file)
        self.assertIn('test_font', self.mc.bitmap_fonts)
        test_font = self.mc.bitmap_fonts['test_font']
        self.assertIsNotNone(test_font)
        self.assertIsNotNone(test_font.bitmap_font)

        self.assertEqual(test_font.bitmap_font.scale_w, 361)
        self.assertEqual(test_font.bitmap_font.scale_h, 512)
        self.assertEqual(test_font.bitmap_font.line_height, 80)
        self.assertEqual(test_font.bitmap_font.base, 57)

        self.assertTrue(len(test_font.bitmap_font.get_characters()), 80)
        self.assertEqual(test_font.bitmap_font.get_characters()[122].id, 122)
        self.assertEqual(test_font.bitmap_font.get_characters()[122].rect["w"], 35)
        self.assertEqual(test_font.bitmap_font.get_characters()[122].rect["h"], 39)
        self.assertEqual(test_font.bitmap_font.get_characters()[122].xadvance, 36)
        self.assertEqual(test_font.bitmap_font.get_characters()[122].xoffset, 1)
        self.assertEqual(test_font.bitmap_font.get_characters()[122].yoffset, 19)
        self.assertEqual(test_font.bitmap_font.get_characters()[122].rect["x"], 58)
        self.assertEqual(test_font.bitmap_font.get_characters()[122].rect["y"], 273)

        # Variable width font (with text descriptor file)
        self.assertIn('test_font_2', self.mc.bitmap_fonts)
        test_font_2 = self.mc.bitmap_fonts['test_font_2']
        self.assertIsNotNone(test_font_2)
        self.assertIsNotNone(test_font_2.bitmap_font)

        self.assertEqual(test_font_2.bitmap_font.scale_w, 330)
        self.assertEqual(test_font_2.bitmap_font.scale_h, 511)
        self.assertEqual(test_font_2.bitmap_font.line_height, 67)
        self.assertEqual(test_font_2.bitmap_font.base, 47)

        self.assertTrue(len(test_font_2.bitmap_font.get_characters()), 80)
        self.assertEqual(test_font_2.bitmap_font.get_characters()[122].id, 122)
        self.assertEqual(test_font_2.bitmap_font.get_characters()[122].rect["w"], 35)
        self.assertEqual(test_font_2.bitmap_font.get_characters()[122].rect["h"], 39)
        self.assertEqual(test_font_2.bitmap_font.get_characters()[122].xadvance, 30)
        self.assertEqual(test_font_2.bitmap_font.get_characters()[122].xoffset, 1)
        self.assertEqual(test_font_2.bitmap_font.get_characters()[122].yoffset, 16)
        self.assertEqual(test_font_2.bitmap_font.get_characters()[122].rect["x"], 80)
        self.assertEqual(test_font_2.bitmap_font.get_characters()[122].rect["y"], 438)

    def test_bitmap_font_text(self):
        # Very basic test
        self.mc.events.post('static_text')
        self.advance_real_time(3)




"""A text widget on a slide."""
import re
from typing import Optional

from kivy.uix.label import Label
from kivy.properties import AliasProperty, NumericProperty, BooleanProperty, \
    ReferenceListProperty, ListProperty
from kivy.graphics import Rectangle, Color, Rotate, Scale

from mpfmc.uix.widget import Widget
from mpfmc.uix.bitmap_font.label_bitmap_font import LabelBitmapFont

MYPY = False
if MYPY:   # pragma: no cover
    from mpfmc.core.mc import MpfMc


# pylint: disable-msg=too-many-instance-attributes
class McFontLabel(Label):

    """Normal label."""

    def get_label(self):
        """Return the label."""
        return self._label


# pylint: disable-msg=too-many-instance-attributes
class BitmapFontLabel(Label):

    """Injects a font or bitmap font into a text widget."""

    def __init__(self, mc: "MpfMc", font_name, **kwargs):

        self.mc = mc
        self.mc.track_leak_reference(self)
        kwargs.setdefault('font_name', font_name)
        kwargs.setdefault('font_kerning', True)

        super().__init__(**kwargs)

    def get_label(self):
        """Return the label."""
        return self._label

    def _create_label(self):
        d = Label._font_properties
        dkw = dict(list(zip(d, [getattr(self, x) for x in d])))
        self._label = LabelBitmapFont(self.mc, **dkw)


var_finder = re.compile(r"(?<=\()[a-zA-Z_0-9|]+(?=\))")
string_finder = re.compile(r"(?<=\$)[a-zA-Z_0-9]+")


class Text(Widget):

    """A text widget on a slide."""

    widget_type_name = 'Text'
    merge_settings = ('font_name', 'font_size', 'bold', 'italic', 'halign',
                      'valign', 'padding_x', 'padding_y', 'text_size',
                      'shorten', 'mipmap', 'markup', 'line_height',
                      'max_lines', 'strip', 'shorten_from', 'split_str',
                      'unicode_errors', 'color', 'casing')
    animation_properties = ('x', 'y', 'font_size', 'color', 'opacity', 'rotation', 'scale')

    def __init__(self, mc: "MpfMc", config: dict, key: Optional[str] = None,
                 play_kwargs: Optional[dict] = None, **kwargs) -> None:
        if 'bitmap_font' in config and config['bitmap_font']:
            if 'font_name' not in config or not config['font_name']:
                raise ValueError("Text widget: font_name is required when bitmap_font is True.")
            self._label = BitmapFontLabel(mc, config['font_name'])
        else:
            self._label = McFontLabel()
        self._label.fbind('texture', self.on_label_texture)
        self.color_instruction = None
        self.rectangle = None
        self.rotate = None
        self.scale_instruction = None

        super().__init__(mc=mc, config=config, key=key)

        # Special handling for baseline anchor
        if self.config['anchor_y'] == 'baseline':
            self.anchor_y = 'bottom'
            self.adjust_bottom = self._label.get_label().get_descent() * -1

        self.original_text = self._get_text_string(config.get('text', ''))

        self.text_variables = dict()
        if play_kwargs:
            self.event_replacements = play_kwargs
        else:
            self.event_replacements = kwargs
        self._process_text(self.original_text)

        # Bind to all properties that when changed need to force
        # the widget to be redrawn
        self.bind(pos=self._draw_widget,
                  size=self._draw_widget,
                  color=self._draw_widget,
                  rotation=self._draw_widget,
                  scale=self._draw_widget)

    def get_text_width(self):
        """Return the text width."""
        if self._label.text:
            # this only makes sense if there is a text
            return self.width
        else:
            # otherwise text width is 0
            return 0

    def __repr__(self) -> str:
        if hasattr(self, '_label') and self._label:
            return '<Text Widget text={}>'.format(self._label.text)
        else:
            return '<Text Widget text=None>'

    def _draw_widget(self, *args):
        """Draws the image (draws a rectangle using the image texture)."""
        del args

        # Redrawing the widget doesn't reposition, so update manually
        anchor = (self.x - self.anchor_offset_pos[0], self.y - self.anchor_offset_pos[1])
        # Save the updated position as a new variable, such that consecutive text
        # changes don't introduce gradual shifts in position.
        pos = self.calculate_rounded_position(anchor)

        if (self._label.text and not self.rectangle) or (self.rectangle and not self._label.text):
            # only create instructions once
            # unfortunately, we also have to redo it when the text becomes empty or non-empty
            self.canvas.clear()
            with self.canvas:
                self.color_instruction = Color(*self.color)
                self.rotate = Rotate(angle=self.rotation, origin=anchor)
                self.scale_instruction = Scale(self.scale)
                self.scale_instruction.origin = anchor

                if self._label.text:
                    self.rectangle = Rectangle(pos=pos, size=self.size, texture=self._label.texture)
                else:
                    self.rectangle = None

        if self.rotate:
            self.rotate.origin = anchor
            self.rotate.angle = self.rotation
        if self.scale_instruction:
            self.scale_instruction.x = self.scale   # Kivy 1.6.0 requires explicit per-axis scale args
            self.scale_instruction.y = self.scale
            self.scale_instruction.z = self.scale
            self.scale_instruction.origin = anchor
        if self.rectangle:
            self.rectangle.pos = pos
            self.rectangle.size = self.size
            self.rectangle.texture = self._label.texture
        if self.color_instruction:
            self.color_instruction.rgba = self.color

    def on_label_texture(self, instance, texture):
        del instance
        if texture:
            self.size = texture.size

            if self.config['anchor_y'] == 'baseline':
                self.adjust_bottom = self._label.get_label().get_descent() * -1

    def update_kwargs(self, **kwargs) -> None:
        self.event_replacements.update(kwargs)
        self._process_text(self.original_text)

    def _get_text_string(self, text: str) -> str:
        if '$' not in text:
            return text

        for text_string in string_finder.findall(text):
            text = text.replace('${}'.format(text_string),
                                self._do_get_text_string(text_string))

        return text

    def _do_get_text_string(self, text_string: str) -> str:
        try:
            return str(self.mc.machine_config['text_strings'][text_string])
        except KeyError:
            # if the text string is not found, put the $ back on
            return '${}'.format(text_string)

    @staticmethod
    def _get_text_vars(text: str):
        return var_finder.findall(text)

    def _process_text(self, text: str) -> None:
        for var_string in self._get_text_vars(text):
            if var_string in self.event_replacements:
                text = text.replace('({})'.format(var_string),
                                    str(self.event_replacements[var_string]))

        if self._get_text_vars(text):
            # monitors won't be added twice, so it's ok to blindly call this
            self._setup_variable_monitors(text)

        self.update_vars_in_text(text)

    def update_vars_in_text(self, text: str) -> None:
        for var_string in self._get_text_vars(text):
            if var_string.startswith('machine|'):
                try:
                    text = text.replace('(' + var_string + ')',
                                        str(self.mc.machine_vars[var_string.split('|')[1]]))
                except KeyError:
                    text = text.replace('(' + var_string + ')', '')

            elif self.mc.player:
                if var_string.startswith('player|'):
                    text = text.replace('(' + var_string + ')',
                                        str(self.mc.player[var_string.split('|')[1]]))
                    continue
                elif var_string.startswith('player') and '|' in var_string:
                    player_num, var_name = var_string.lstrip('player').split('|')
                    try:
                        value = self.mc.player_list[int(player_num) - 1][
                            var_name]

                        if value is not None:
                            text = text.replace('(' + var_string + ')',
                                                str(value))
                        else:
                            text = text.replace('(' + var_string + ')', '')
                    except IndexError:
                        text = text.replace('(' + var_string + ')', '')
                    continue
                elif self.mc.player.is_player_var(var_string):
                    text = text.replace('(' + var_string + ')',
                                        str(self.mc.player[var_string]))
                    continue

            if var_string in self.event_replacements:
                text = text.replace('({})'.format(var_string),
                                    str(self.event_replacements[var_string]))

        self.update_text(text)

    def update_text(self, text: str) -> None:
        if text:
            if self.config['min_digits']:
                text = text.zfill(self.config['min_digits'])

            if self.config['number_grouping']:

                # find the numbers in the string
                number_list = [s for s in text.split() if s.isdigit()]

                # group the numbers and replace them in the string
                for item in number_list:
                    grouped_item = Text.group_digits(item)
                    text = text.replace(str(item), grouped_item)

            if self.config.get('casing', None) in ('lower', 'upper', 'title', 'capitalize'):
                text = getattr(text, self.config['casing'])()

        self._label.text = text
        self._label.texture_update()
        self._draw_widget()

    def _player_var_change(self, **kwargs) -> None:
        del kwargs
        self.update_vars_in_text(self.original_text)

    def _machine_var_change(self, **kwargs) -> None:
        del kwargs
        self.update_vars_in_text(self.original_text)

    def _setup_variable_monitors(self, text: str) -> None:
        for var_string in self._get_text_vars(text):
            if '|' not in var_string:
                self.add_player_var_handler(name=var_string)
                self.add_current_player_handler()
            else:
                source, variable_name = var_string.split('|')
                if source.lower().startswith('player'):

                    if source.lstrip('player'):  # we have player num
                        self.add_player_var_handler(name=variable_name)
                    else:  # no player num
                        self.add_player_var_handler(name=variable_name)
                        self.add_current_player_handler()
                elif source.lower() == 'machine':
                    self.add_machine_var_handler(name=variable_name)

    def add_player_var_handler(self, name: str) -> None:
        self.mc.events.replace_handler('player_{}'.format(name),
                                       self._player_var_change)

    def add_current_player_handler(self) -> None:
        self.mc.events.replace_handler('player_turn_start',
                                       self._player_var_change)

    def add_machine_var_handler(self, name: str) -> None:
        self.mc.events.add_handler('machine_var_{}'.format(name),
                                   self._machine_var_change)

    def prepare_for_removal(self) -> None:
        super().prepare_for_removal()
        self.mc.events.remove_handler(self._player_var_change)
        self.mc.events.remove_handler(self._machine_var_change)

    @staticmethod
    def group_digits(text: str, separator: str = ',', group_size: int = 3) -> str:
        """Enable digit grouping (i.e. adds comma separators between thousands digits).

        Args:
            text: The incoming string of text
            separator: String of the character(s) you'd like to add between the
                digit groups. Default is a comma. (",")
            group_size: How many digits you want in each group. Default is 3.

        Returns: A string with the separator added.

        MPF uses this method instead of the Python locale settings because the
        locale settings are a mess. They're set system-wide and it's really
        hard
        to make them work cross-platform and there are all sorts of external
        dependencies, so this is just way easier.

        """
        digit_list = list(text.split('.')[0])

        for i in range(len(digit_list))[::-group_size][1:]:
            digit_list.insert(i + 1, separator)

        return ''.join(digit_list)

    #
    # Properties
    #

    disabled_color = ListProperty([1, 1, 1, .3])
    '''The color of the text when the widget is disabled, in the (r, g, b, a)
    format.

    .. versionadded:: 1.8.0

    :attr:`disabled_color` is a :class:`~kivy.properties.ListProperty` and
    defaults to [1, 1, 1, .3].
    '''

    def _get_text(self) -> str:
        return self._label.text

    def _set_text(self, text: str) -> None:
        self._label.text = text

    text = AliasProperty(_get_text, _set_text)
    '''Text of the label.
    '''

    def _get_text_size(self) -> list:
        return self._label.text_size

    def _set_text_size(self, text_size: list) -> None:
        self._label.text_size = text_size

    text_size = AliasProperty(_get_text_size, _set_text_size)
    '''By default, the label is not constrained to any bounding box.
    You can set the size constraint of the label with this property.
    The text will autoflow into the constraints. So although the font size
    will not be reduced, the text will be arranged to fit into the box as best
    as possible, with any text still outside the box clipped.

    This sets and clips :attr:`texture_size` to text_size if not None.

    .. versionadded:: 1.0.4

    For example, whatever your current widget size is, if you want the label to
    be created in a box with width=200 and unlimited height::

        Label(text='Very big big line', text_size=(200, None))

    .. note::

        This text_size property is the same as the
        :attr:`~kivy.core.text.Label.usersize` property in the
        :class:`~kivy.core.text.Label` class. (It is named size= in the
        constructor.)

    :attr:`text_size` is a :class:`~kivy.properties.ListProperty` and
    defaults to (None, None), meaning no size restriction by default.
    '''

    def _get_font_name(self) -> str:
        return self._label.font_name

    def _set_font_name(self, font_name: str) -> None:
        self._label.font_name = font_name

    font_name = AliasProperty(_get_font_name, _set_font_name)
    '''Filename of the font to use. The path can be absolute or relative.
    Relative paths are resolved by the :func:`~kivy.resources.resource_find`
    function.

    .. warning::

        Depending of your text provider, the font file can be ignored. However,
        you can mostly use this without problems.

        If the font used lacks the glyphs for the particular language/symbols
        you are using, you will see '[]' blank box characters instead of the
        actual glyphs. The solution is to use a font that has the glyphs you
        need to display. For example, to display |unicodechar|, use a font such
        as freesans.ttf that has the glyph.

        .. |unicodechar| image:: images/unicode-char.png

    :attr:`font_name` is a :class:`~kivy.properties.StringProperty` and
    defaults to 'Roboto'. This value is taken
    from :class:`~kivy.config.Config`.
    '''

    def _get_font_size(self):
        return self._label.font_size

    def _set_font_size(self, font_size) -> None:
        self._label.font_size = font_size

    font_size = AliasProperty(_get_font_size, _set_font_size)
    '''Font size of the text, in pixels.

    :attr:`font_size` is a :class:`~kivy.properties.NumericProperty` and
    defaults to 15sp.
    '''

    def _get_line_height(self) -> float:
        return self._label.line_height

    def _set_line_height(self, line_height: float) -> None:
        self._label.line_height = line_height

    bitmap_font = BooleanProperty(False)
    '''Flag indicating whether or not the font_name attribute refers to a
    bitmap font.'''

    line_height = AliasProperty(_get_line_height, _set_line_height)
    '''Line Height for the text. e.g. line_height = 2 will cause the spacing
    between lines to be twice the size.

    :attr:`line_height` is a :class:`~kivy.properties.NumericProperty` and
    defaults to 1.0.

    .. versionadded:: 1.5.0
    '''

    def _get_bold(self) -> bool:
        return self._label.bold

    def _set_bold(self, bold: bool) -> None:
        self._label.bold = bold

    bold = AliasProperty(_get_bold, _set_bold)
    '''Indicates use of the bold version of your font.

    .. note::

        Depending of your font, the bold attribute may have no impact on your
        text rendering.

    :attr:`bold` is a :class:`~kivy.properties.BooleanProperty` and defaults to
    False.
    '''

    def _get_italic(self) -> bool:
        return self._label.italic

    def _set_italic(self, italic: bool) -> None:
        self._label.italic = italic

    italic = AliasProperty(_get_italic, _set_italic)
    '''Indicates use of the italic version of your font.

    .. note::

        Depending of your font, the italic attribute may have no impact on your
        text rendering.

    :attr:`italic` is a :class:`~kivy.properties.BooleanProperty` and defaults
    to False.
    '''

    def _get_underline(self) -> bool:
        return self._label.underline

    def _set_underline(self, underline: bool) -> None:
        self._label.underline = underline

    underline = AliasProperty(_get_underline, _set_underline)
    '''Adds an underline to the text.

    .. note::
        This feature requires the SDL2 text provider.

    .. versionadded:: 1.10.0

    :attr:`underline` is a :class:`~kivy.properties.BooleanProperty` and
    defaults to False.
    '''

    def _get_strikethrough(self) -> bool:
        return self._label.strikethrough

    def _set_strikethrough(self, strikethrough: bool) -> None:
        self._label.strikethrough = strikethrough

    strikethrough = AliasProperty(_get_strikethrough, _set_strikethrough)
    '''Adds a strikethrough line to the text.

    .. note::
        This feature requires the SDL2 text provider.
    '''

    def _get_padding_x(self):
        return self._label.padding_x

    def _set_padding_x(self, padding_x):
        self._label.padding_x = padding_x

    padding_x = AliasProperty(_get_padding_x, _set_padding_x)
    '''Horizontal padding of the text inside the widget box.
    '''

    def _get_padding_y(self):
        return self._label.padding_y

    def _set_padding_y(self, padding_y):
        self._label.padding_y = padding_y

    padding_y = AliasProperty(_get_padding_y, _set_padding_y)
    '''Vertical padding of the text inside the widget box.
    '''

    padding = ReferenceListProperty(padding_x, padding_y)
    '''Padding of the text in the format (padding_x, padding_y)

    :attr:`padding` is a :class:`~kivy.properties.ReferenceListProperty` of
    (:attr:`padding_x`, :attr:`padding_y`) properties.
    '''

    def _get_halign(self) -> str:
        return self._label.halign

    def _set_halign(self, halign: str) -> None:
        self._label.halign = halign

    halign = AliasProperty(_get_halign, _set_halign)
    '''Horizontal alignment of the text. Available options are : left, center,
    right and justify.

    .. warning::

        This doesn't change the position of the text texture of the Label
        (centered), only the position of the text in this texture. You probably
        want to bind the size of the Label to the :attr:`texture_size` or set a
        :attr:`text_size`.
    '''

    def _get_valign(self) -> str:
        return self._label.valign

    def _set_valign(self, valign: str) -> None:
        self._label.valign = valign

    valign = AliasProperty(_get_valign, _set_valign)
    '''Vertical alignment of the text. Available options are : `'bottom'`,
    `'middle'` (or `'center'`) and `'top'`.

    .. warning::

        This doesn't change the position of the text texture of the Label
        (centered), only the position of the text within this texture. You
        probably want to bind the size of the Label to the :attr:`texture_size`
        or set a :attr:`text_size` to change this behavior.
    '''

    def _get_outline_width(self) -> Optional[int]:
        return self._label.outline_width

    def _set_outline_width(self, outline_width: Optional[int]) -> None:
        self._label.outline_width = outline_width

    outline_width = AliasProperty(_get_outline_width, _set_outline_width)
    '''Width in pixels for the outline around the text. No outline will be
    rendered if the value is None.

    .. note::
        This feature requires the SDL2 text provider.
    '''

    def _get_outline_color(self) -> list:
        return self._label.outline_color

    def _set_outline_color(self, outline_color: list) -> None:
        self._label.outline_color = outline_color

    outline_color = AliasProperty(_get_outline_color, _set_outline_color)
    '''The color of the text outline, in the (r, g, b) format.

    .. note::
        This feature requires the SDL2 text provider.
    '''

    def _get_disabled_outline_color(self) -> list:
        return self._label.disabled_outline_color

    def _set_disabled_outline_color(self, disabled_outline_color: list) -> None:
        self._label.disabled_outline_color = disabled_outline_color

    disabled_outline_color = AliasProperty(_get_disabled_outline_color, _set_disabled_outline_color)
    '''The color of the text outline when the widget is disabled, in the
    (r, g, b) format.

    .. note::
        This feature requires the SDL2 text provider.
    '''

    def _get_mipmap(self) -> bool:
        return self._label.mipmap

    def _set_mipmap(self, mipmap: bool) -> None:
        self._label.mipmap = mipmap

    mipmap = AliasProperty(_get_mipmap, _set_mipmap)
    '''Indicates whether OpenGL mipmapping is applied to the texture or not.
    Read :ref:`mipmap` for more information.
    '''

    def _get_shorten(self) -> bool:
        return self._label.shorten

    def _set_shorten(self, shorten: bool) -> None:
        self._label.shorten = shorten

    shorten = AliasProperty(_get_shorten, _set_shorten)
    '''
    Indicates whether the label should attempt to shorten its textual contents
    as much as possible if a :attr:`text_size` is given. Setting this to True
    without an appropriately set :attr:`text_size` will lead to unexpected
    results.

    :attr:`shorten_from` and :attr:`split_str` control the direction from
    which the :attr:`text` is split, as well as where in the :attr:`text` we
    are allowed to split.
    '''

    def _get_shorten_from(self) -> str:
        return self._label.shorten_from

    def _set_shorten_from(self, shorten_from: str) -> None:
        self._label.shorten_from = shorten_from

    shorten_from = AliasProperty(_get_shorten_from, _set_shorten_from)
    '''The side from which we should shorten the text from, can be left,
    right, or center.

    For example, if left, the ellipsis will appear towards the left side and we
    will display as much text starting from the right as possible. Similar to
    :attr:`shorten`, this option only applies when :attr:`text_size` [0] is
    not None, In this case, the string is shortened to fit within the specified
    width.
    '''

    def _get_is_shortened(self) -> bool:
        return self._label.is_shortened

    is_shortened = AliasProperty(_get_is_shortened, None)
    '''This property indicates if :attr:`text` was rendered with or without
    shortening when :attr:`shorten` is True.
    '''

    def _get_split_str(self) -> str:
        return self._label.split_str

    def _set_split_str(self, split_str: str) -> None:
        self._label.split_str = split_str

    split_str = AliasProperty(_get_split_str, _set_split_str)
    '''The string used to split the :attr:`text` while shortening the string
    when :attr:`shorten` is True.

    For example, if it's a space, the string will be broken into words and as
    many whole words that can fit into a single line will be displayed. If
    :attr:`split_str` is the empty string, `''`, we split on every character
    fitting as much text as possible into the line.
    '''

    def _get_ellipsis_options(self) -> dict:
        return self._label.ellipsis_options

    def _set_ellipsis_options(self, ellipsis_options: dict) -> None:
        self._label.ellipsis_options = ellipsis_options

    ellipsis_options = AliasProperty(_get_ellipsis_options, _set_ellipsis_options)
    '''Font options for the ellipsis string('...') used to split the text.

    Accepts a dict as option name with the value. Only applied when
    :attr:`markup` is true and text is shortened. All font options which work
    for :class:`Label` will work for :attr:`ellipsis_options`. Defaults for
    the options not specified are taken from the surronding text.

    .. code-block:: kv

        Label:
            text: 'Some very long line which will be cut'
            markup: True
            shorten: True
            ellipsis_options: {'color':(1,0.5,0.5,1),'underline':True}
    '''

    def _get_unicode_errors(self) -> str:
        return self._label.unicode_errors

    def _set_unicode_errors(self, unicode_errors: str) -> None:
        self._label.unicode_errors = unicode_errors

    unicode_errors = AliasProperty(_get_unicode_errors, _set_unicode_errors)
    '''How to handle unicode decode errors. Can be `'strict'`, `'replace'` or
    `'ignore'`.
    '''

    def _get_markup(self) -> bool:
        return self._label.markup

    def _set_markup(self, markup: bool) -> None:
        self._label.markup = markup

    markup = AliasProperty(_get_markup, _set_markup)
    '''If True, the text will be rendered using the
    :class:`~kivy.core.text.markup.MarkupLabel`: you can change the
    style of the text using tags. Check the
    :doc:`api-kivy.core.text.markup` documentation for more information.
    '''

    def _get_refs(self) -> dict:
        return self._label.refs

    refs = AliasProperty(_get_refs, None)
    '''List of ``[ref=xxx]`` markup items in the text with the bounding box of
    all the words contained in a ref, available only after rendering.

    For example, if you wrote::

        Check out my [ref=hello]link[/ref]

    The refs will be set with::

        {'hello': ((64, 0, 78, 16), )}

    The references marked "hello" have a bounding box at (x1, y1, x2, y2).
    These co-ordinates are relative to the top left corner of the text, with
    the y value increasing downwards. You can define multiple refs with the
    same name: each occurrence will be added as another (x1, y1, x2, y2) tuple
    to this list.

    The current Label implementation uses these references if they exist in
    your markup text, automatically doing the collision with the touch and
    dispatching an `on_ref_press` event.

    You can bind a ref event like this::

        def print_it(instance, value):
            print('User click on', value)
        widget = Label(text='Hello [ref=world]World[/ref]', markup=True)
        widget.on_ref_press(print_it)

    .. note::

        This works only with markup text. You need :attr:`markup` set to
        True.
    '''

    def _get_max_lines(self) -> Optional[int]:
        return self._label.max_lines

    def _set_max_lines(self, max_lines: Optional[int]) -> None:
        self._label.max_lines = max_lines

    max_lines = AliasProperty(_get_max_lines, _set_max_lines)
    '''Maximum number of lines to use, defaults to 0, which means unlimited.
    Please note that :attr:`shorten` take over this property. (with
    shorten, the text is always one line.)
    '''

    def _get_strip(self) -> bool:
        return self._label.strip

    def _set_strip(self, strip: bool) -> None:
        self._label.strip = strip

    strip = AliasProperty(_get_strip, _set_strip)
    '''Whether leading and trailing spaces and newlines should be stripped from
    each displayed line. If True, every line will start at the right or left
    edge, depending on :attr:`halign`. If :attr:`halign` is `justify` it is
    implicitly True.

    .. versionadded:: 1.9.0

    :attr:`strip` is a :class:`~kivy.properties.BooleanProperty` and
    defaults to False.
    '''

    def _get_font_hinting(self) -> Optional[str]:
        return self._label.font_hinting

    def _set_font_hinting(self, font_hinting: Optional[str]):
        self._label.font_hinting = font_hinting

    font_hinting = AliasProperty(_get_font_hinting, _set_font_hinting)
    '''What hinting option to use for font rendering.
    Can be one of `'normal'`, `'light'`, `'mono'` or None.

    .. note::
        This feature requires the SDL2 text provider.
    '''

    def _get_font_kerning(self) -> bool:
        return self._label.font_kerning

    def _set_font_kerning(self, font_kerning: bool) -> None:
        self._label.font_kerning = font_kerning

    font_kerning = AliasProperty(_get_font_kerning, _set_font_kerning)
    '''Whether kerning is enabled for font rendering.

    .. note::
        This feature requires the SDL2 text provider.
    '''

    def _get_font_blended(self) -> bool:
        return self._label.font_blended

    def _set_font_blended(self, font_blended: bool) -> None:
        self._label.font_blended = font_blended

    font_blended = AliasProperty(_get_font_blended, _set_font_blended)
    '''Whether blended or solid font rendering should be used.

    .. note::
        This feature requires the SDL2 text provider.
    '''

    rotation = NumericProperty(0)
    '''Rotation angle value of the widget.

    :attr:`rotation` is an :class:`~kivy.properties.NumericProperty` and defaults to
    0.
    '''

    scale = NumericProperty(1.0)
    '''Scale value of the widget.

    :attr:`scale` is an :class:`~kivy.properties.NumericProperty` and defaults to
    1.0.
    '''


widget_classes = [Text]



"""A widgets showing a camera image."""
from kivy.uix.camera import Camera
from mpfmc.uix.widget import Widget


class CameraWidget(Widget, Camera):

    """A widgets showing a camera image."""

    widget_type_name = "Camera"

    def __init__(self, mc, config, key=None, **kwargs):
        super().__init__(mc=mc, config=config, key=key)

    def pass_to_kivy_widget_init(self):
        return dict(resolution=(self.config['width'],
                                self.config['height']),
                    index=self.config['camera_index'])


widget_classes = [CameraWidget]
